#ifndef SELDON_FILE_FACTORIZATION_LU_CXX

namespace Seldon
{
  
#ifndef SELDON_WITH_LAPACK
  //! LU factorisation of indefinite unsymmetric matrices with partial pivoting
  /*!
    \param[inout] A on input, original matrix; on output, LU factors
    \param[inout] IPivot pivot numbers
    \param[in] exemple_reel any real number
    \param[in] zero zero (real or complex depending on the type of T)
    \param[in] one one (real or complex depending on the type of T)
   */
  template<class T, class Storage, class Allocator>
  void GetLU(Matrix<T, General, Storage, Allocator>& A, IVect& IPivot,
	     LapackInfo& info_)
  {
    T zero, one; SetComplexZero(zero); SetComplexOne(one);
    int m = A.GetM();
    int n = A.GetN();
    IPivot.Reallocate(n);
    
    // Quick return if possible
    if ((m==0)||(n==0)) 
      return;
    
    typename ClassComplexType<T>::Treal modulus_coef, tmp_real;
    T tmp_cpx; int jp, info=0;
    
    for (int j = 0; j < min(m,n); j++)
      {
	// we search maximum of the column j, it is the pivot
	jp = j; modulus_coef = abs(A(j,j));
	for (int k = j+1; k < m; k++)
	  {
	    tmp_real = abs(A(k, j));
	    if (tmp_real > modulus_coef)
	      {
		jp = k;
		modulus_coef = tmp_real;
	      }
	  }
	
	IPivot(j) = jp+1;
	if (A(jp,j) != zero)
	  {
	    // we don't have singularity
	    // we swap row j with row jp
	    if (jp != j)
	      {
		for (int k = 0; k < n; k++)
		  {
		    tmp_cpx = A(j, k);
		    A(j, k) = A(jp, k);
		    A(jp, k) = tmp_cpx;
		  }
	      }
	    
	    // we put coefficients of Lower Matrix L
	    if (j < m)
	      { 
		tmp_cpx = one/A(j, j);
		for (int k = j+1; k < m; k++)
		  A(k, j) *= tmp_cpx;
		
		A(j, j) = tmp_cpx;
	      }
	  }
	else if (info==0)
	  {
	    // singularity on row j
	    info = j+1;
	    cout<<"Matrix is singular"<<endl;
	    abort();	    
	  }
	
	// we modify row j+1 until m 
	if (j < min(m, n))
	  {
	    for (int k = j+1; k < m; k++)
	      for (int i = j+1; i < n; i++)
		A(k, i) -= A(k, j)*A(j, i);
	  }
      }
  }
  
  
  //! Resolution of A x = b, by using LU factorisation of A
  /*!
    \param[in] TransA if SeldonNoTrans, resolution of A x = b
                      if SeldonTrans, resolution of A^T x = b
    \param[in] A factorisation of matrix (generated by GetLU)
    \param[in] IPivot pivot numbers (generated by GetLU)
    \param[inout] b on input, right hand side; on output, solution of A x = b
   */
  template<class T, class Allocator,
           class Storage, class Prop, class Allocator2>
  void SolveLuVector(const SeldonTranspose& TransA,
                     const Matrix<T, General, Storage, Allocator>& A,
                     const IVect& IPivot, Vector<T, Prop, Allocator2>& b,
		     LapackInfo& info_)
  {
    int n = A.GetM();
    if (n <= 0)
      return;
    
    if (TransA.Trans())
      {
        int l; T tmp;
        // we solve now U^T x = b
        // by columns
        for (int k = 0; k < n; k++)
          {
            for (l = 0; l < k; l++)
              b(k) -= A(l, k)*b(l);
            
            b(k) *= A(k, k);
          }
        
        // we solve L^T x = b 
        // by columns
        for (int k = n-2; k >= 0; k--)
          {
            for (l = (k+1); l < n; l++)
              b(k) -= A(l, k)*b(l); 
          }     

        for (int k = n-2; k >= 0; k--)
          {
            l = IPivot(k)-1;
            tmp = b(l);
            // we swap if necessary
            if (l != k)
              {
                b(l) = b(k);
                b(k) = tmp;
              }
          }
      }
    else
      {
        int l; T tmp;
        // we solve Lx = b 
        // by columns
        for (int k = 0; k < (n-1); k++)
          {
            l = IPivot(k)-1;
            tmp = b(l);
            // we swap if necessary
            if (l != k)
              {
                b(l) = b(k);
                b(k) = tmp;
              }
          }
        
        for (int k = 0; k < (n-1); k++)
          {
            tmp = b(k);
            for ( l = (k+1); l < n; l++)
              b(l) -= A(l, k)*tmp; 
          }
        
        // we solve now Ux = b
        // by columns
        for (int k = (n-1); k >= 0; k--)
          {
            b(k) *= A(k, k);
            tmp = b(k);
            for ( l = 0; l < k; l++)
              b(l) -= A(l, k)*tmp;
          }
      }
  }
    
  
  //! Resolution of A x = b, by using LU factorisation of A
  /*!
    \param[in] A factorisation of matrix (generated by GetLU)
    \param[in] IPivot pivot numbers (generated by GetLU)
    \param[inout] b on input, right hand side; on output, solution of A x = b
   */
  template<class T, class Allocator,
           class Storage, class Prop, class Allocator2>
  void SolveLuVector(const Matrix<T, General, Storage, Allocator>& A,
                     const IVect& IPivot, Vector<T, Prop, Allocator2>& b,
		     LapackInfo& info)
  {
    SolveLU(SeldonNoTrans, A, IPivot, b);
  }
  
  
  //! LU factorisation of indefinite unsymmetric matrices with partial pivoting
  /*!
    \param[inout] A on input, original matrix; on output, LU factors
    \param[inout] IPivot pivot numbers
    \param[in] exemple_reel any real number
    \param[in] zero zero (real or complex depending on the type of T)
    \param[in] one one (real or complex depending on the type of T)
    we compute a LDL^t factorization with partial symmetric pivoting
    D is a block diagonal matrix with 2x2 and 1x1 blocks
   */
  template<class T, class Storage, class Allocator>
  void GetLU(Matrix<T, Symmetric, Storage, Allocator>& A, IVect& IPivot,
	     LapackInfo& info_)   
  {
    T zero, one; SetComplexZero(zero); SetComplexOne(one);
    int n = A.GetN();
    IPivot.Reallocate(n);
    // Quick return if possible
    if (n == 0) 
      return;
    
    typedef typename ClassComplexType<T>::Treal reel;
    typename ClassComplexType<T>::Treal modulus_coef, rowmax, colmax, alpha, sevten;
    sevten = reel(17);
    alpha = (reel(1)+sqrt(sevten))/8;
    T tmp_cpx, d11, d22, d21, wk, wkp1;
    reel zero_r(0); 
    int info = 0;
    
    // k is the main loop index, increasing from 1 to n in steps of
    // 1 or 2
    int i,k = 0, j = 0, kc = 0,knc, npp = A.GetDataSize();
    int kstep , kp, kx, kpc, kk, imax(0); 
    T* a_ptr = A.GetData();
    while (k < n)
      {
	knc = kc; kstep = 1;
	// Determine rows and columns to be interchanged and whether
	// a 1-by-1 or 2-by-2 pivot block will be used
	modulus_coef = abs(a_ptr[kc]);
	// IMAX is the row-index of the largest off-diagonal element in
	// column K, and COLMAX is its absolute value
	if (k < (n-1))
	  {
	    colmax = zero_r;
	    for ( j = 1; j < (n-k); j++)
	      if (abs(a_ptr[kc+j]) > colmax)
		{
		  imax = k + j;
		  colmax = abs(a_ptr[kc+j]);
		}
	  }
	else
	  colmax = zero_r;
	
	if (max(modulus_coef,colmax)==zero_r)
	  {
	    // Column K is zero: set INFO and continue
	    if (info==0)
	      info = k+1;
	    kp = k;
	  }
	else
	  {
	    if (modulus_coef >= (alpha*colmax))
	      {
		// no interchange, use 1-by-1 pivot block
		kp = k;
	      }
	    else
	      {
		// JMAX is the column-index of the largest off-diagonal
		// element in row IMAX, and ROWMAX is its absolute value
		rowmax = zero_r;
		kx = kc + imax - k;
		for ( j=k ; j<=(imax-1) ; j++)
		  {
		    if (abs(a_ptr[kx]) > rowmax)
		      rowmax = abs(a_ptr[kx]);
		
		    kx += n-1-j;
		  }
		kpc = npp - (n-imax)*(n-imax+1) / 2;
		if (imax<(n-1))
		  {
		    for ( j = 1; j < (n-imax); j++)
		      {
			if (abs(a_ptr[kpc+j]) > rowmax)
			  rowmax = abs(a_ptr[kpc+j]);
		      }
		  }
		if (modulus_coef >= (alpha*colmax*(colmax/rowmax)) )
		  {
		    // no interchange, use 1-by-1 pivot block
		    kp = k;
		  }
		else if (abs(a_ptr[kpc]) >= (alpha*rowmax) )
		  {
		    // interchange rows and columns K and IMAX, use 1-by-1
		    // pivot block
		    kp = imax;
		  }
		else
		  {
		    // interchange rows and columns K+1 and IMAX, use 2-by-2
		    // pivot block
		    kp = imax;
		    kstep = 2;
		  }    
	      }
	    kk = k + kstep - 1;
	    if (kstep==2)
	      knc += n-k;
	    if ( kp != kk )
	      {
		// Interchange rows and columns KK and KP in the trailing
		// submatrix A(k:n,k:n)
		if (kp<(n-1))
		  {
		    for ( j = 1; j < (n-kp); j++)
		      {
			tmp_cpx = a_ptr[kpc+j];
			a_ptr[kpc+j] = a_ptr[knc+kp-kk+j];
			a_ptr[knc+kp-kk+j] = tmp_cpx;
		      }
		  }
		kx = knc+kp-kk;
		for ( j = (kk+1); j < kp; j++)
		  {
		    kx += n-j;
		    tmp_cpx = a_ptr[knc+j-kk];
		    a_ptr[knc+j-kk] = a_ptr[kx];
		    a_ptr[kx] = tmp_cpx;
		  }
		tmp_cpx = a_ptr[knc];
		a_ptr[knc] = a_ptr[kpc];
		a_ptr[kpc] = tmp_cpx;
		if (kstep == 2)
		  {
		    tmp_cpx = a_ptr[kc+1];
		    a_ptr[kc+1] = a_ptr[kc+kp-k];
		    a_ptr[kc+kp-k] = tmp_cpx;
		  }
	      }
	    // Update the trailing submatrix
	    if (kstep == 1)
	      {
		// 1-by-1 pivot block D(k): column k now holds
		// W(k) = L(k)*D(k)
		// where L(k) is the k-th column of L
		if (k < (n-1))
		  {
		    // Perform a rank-1 update of A(k+1:n,k+1:n) as
		    // A := A - L(k)*D(k)*L(k)' = A - W(k)*(1/D(k))*W(k)'
		    d11 = one/a_ptr[kc];
		    
		    kk = kc+n-k;
		    for ( j = 1; j < (n-k); j++)
		      {
			if (a_ptr[kc+j] != zero )
			  {
			    tmp_cpx = d11*a_ptr[kc+j];
			    int l = kk;
			    for ( i = j; i < (n-k); i++)
			      a_ptr[l++] -= a_ptr[kc+i]*tmp_cpx;
			  }
			kk += n - k - j;
		      }
		    
		    // Store L(k) in column K
		    for ( j = 1; j < (n-k); j++)
		      a_ptr[kc+j] *= d11;
		  }
	      }
	    else
	      {
		// 2-by-2 pivot block D(k): columns K and K+1 now hold
		// ( W(k) W(k+1) ) = ( L(k) L(k+1) )*D(k)
		// where L(k) and L(k+1) are the k-th and (k+1)-th columns
		// of L
		if ( k < (n-2) )
		  {
		    // Perform a rank-2 update of A(k+2:n,k+2:n) as
		    // A := A - ( L(k) L(k+1) )*D(k)*( L(k) L(k+1) )'
		    //    = A - ( W(k) W(k+1) )*inv(D(k))*( W(k) W(k+1) )'
		    // where L(k) and L(k+1) are the k-th and (k+1)-th
		    // columns of L
		    d21 = a_ptr[k+1+k*(2*n-k-1)/2];
		    d11 = a_ptr[k+1+(k+1)*(2*n-k-2)/2]/d21;
		    d22 = a_ptr[k+k*(2*n-k-1)/2]/d21;
		    tmp_cpx = one/(d11*d22-one);
		    d21 = tmp_cpx/d21;
		    for ( j = (k+2); j < n; j++)
		      {
			wk = d21*( d11*a_ptr[ j + k*(2*n-k-1)/2 ]
                                   - a_ptr[ j + (k+1)*(2*n-k-2)/2 ] );
			
                        wkp1 = d21*( d22*a_ptr[ j + (k+1)*(2*n-k-2)/2 ]
                                     - a_ptr[ j + k*(2*n-k-1)/2 ] );
			
                        for ( i = j; i < n; i++)
                          a_ptr[i+j*(2*n-j-1)/2] -= a_ptr[i+k*(2*n-k-1)/2 ]*wk
                            + a_ptr[i+(k+1)*(2*n-k-2)/2 ]*wkp1;

			a_ptr[j + k*(2*n-k-1)/2 ] = wk;
			a_ptr[j + (k+1)*(2*n-k-2)/2 ] = wkp1;
		      }
		  }
	      }
	  }
        
	// Store details of the interchanges in IPivot
	if (kstep==1)
	  IPivot(k) = kp+1;
	else
	  {
	    IPivot(k) = -(kp+1);
	    IPivot(k+1) = -(kp+1);
	  }
        
	// Increase K and return to the start of the main loop
	k += kstep;
	kc = knc + n - k + 1;
      }
  }
  
  
  //! Resolution of A x = b, by using LU factorisation of A
  /*!
    \param[in] A factorisation of matrix (generated by GetLU)
    \param[in] IPivot pivot numbers (generated by GetLU)
    \param[inout] b on input, right hand side; on output, solution of A x = b
    This method is written for symmetric matrices (real or complex)
   */
  template<class T, class Allocator, class Storage, class Prop, class Allocator2>
  void SolveLuVector(const Matrix<T, Symmetric, Storage, Allocator>& A,
                     const IVect& IPivot, Vector<T, Prop, Allocator2>& b,
		     LapackInfo& info)
  {
    int n = A.GetM();
    if (n <= 0)
      return;
    
    // Solve A*X = B, where A = L*D*L'.
    // First solve L*D*X = B, overwriting B with X.
    
    // k is the main loop index, increasing from 1 to n in steps of
    // 1 or 2, depending on the size of the diagonal blocks.
    int k = 0, kc = 0, j = 0, kp;
    T tmp, akm1k, akm1, ak, denom, bkm1, bk, one;
    SetComplexOne(one);
    T* a_ptr = A.GetData();
    while (k < n)
      {
	if (IPivot(k) > 0)
	  {
	    // 1 x 1 diagonal block
	    //
	    // Interchange rows k and IPivot(k).
	    kp = IPivot(k) - 1;
	    if (kp != k)
	      {
		tmp = b(k);
		b(k) = b(kp);
		b(kp) = tmp;
	      }
	    
	    if (k < (n-1))
	      {
		for ( j = 1; j < (n-k); j++)
		  b(k+j) -= a_ptr[kc+j]*b(k);
	      }
	    b(k) /= a_ptr[kc];
	    kc += n-k;
	    k++;
	  }
	else
	  {
	    // 2 x 2 diagonal block
	    //
	    // Interchange rows k+1 and -IPivot(k).
	    kp = -IPivot(k) - 1;
	    if (kp != (k+1))
	      {
		tmp = b(k+1);
		b(k+1) = b(kp);
		b(kp) = tmp;
	      }
	    if (k<(n-1))
	      {
		for ( j = 1; j < (n-k-1); j++)
		  b(k+j+1) -= a_ptr[kc+j+1]*b(k);
		
		for ( j = 1; j < (n-k-1); j++)
		  b(k+j+1) -= a_ptr[kc+n-k+j]*b(k+1);
	      }
	    akm1k = a_ptr[kc+1];
	    akm1 = a_ptr[kc]/akm1k;
	    ak = a_ptr[kc+n-k]/akm1k;
	    denom = akm1*ak - one;
	    bkm1 = b(k)/akm1k;
	    bk = b(k+1)/akm1k;
	    b(k) = (ak*bkm1-bk)/denom;
	    b(k+1) = (akm1*bk-bkm1)/denom;
	    kc += 2*(n-k-1) + 1;
	    k += 2;
	  }
	
      }
    k = n-1;
    kc = A.GetDataSize();
    // Next solve L'*X = B, overwriting B with X.
    
    // k is the main loop index, decreasing from n to 1 in steps of
    // 1 or 2, depending on the size of the diagonal blocks.
    
    while (k >= 0)
      {
	kc -= n-k;
	if (IPivot(k) > 0)
	  {
	    // 1 x 1 diagonal block
	    
	    // Multiply by inv(L'(K)), where L(K) is the transformation
	    // stored in column K of A.
	    if (k < (n-1))
	      {
		for ( j = 1; j < (n-k); j++)
		  b(k) -= a_ptr[kc+j]*b(k+j);
	      }
	    kp = IPivot(k) - 1;
	    if (kp != k)
	      {
		tmp = b(k);
		b(k) = b(kp);
		b(kp) = tmp;
	      }
	    k--;
	  }
	else
	  {
	    // 2 x 2 diagonal block
	    
	    // Multiply by inv(L'(K-1)), where L(K-1) is the transformation
	    // stored in columns K-1 and K of A.
	    if (k < (n-1))
	      {
		for ( j = 1; j < (n-k); j++)
		  b(k) -= a_ptr[kc+j]*b(k+j);
		
		kp = kc-(n-k);
		for ( j = 1; j < (n-k); j++)
		  b(k-1) -= a_ptr[kp+j]*b(k+j);
	      }
	    // Interchange rows K and -IPIV(K).
	    kp = -IPivot(k)-1;
	    if (kp != k)
	      {
		tmp = b(k);
		b(k) = b(kp);
		b(kp) = tmp;
	      }
	    kc -= (n-k+1);
	    k -= 2;
	  }
      }
  }
  

  //! A is replaced by its inverse
  /*!
    \param[inout] A on input, original matrix; on output, inverse
    The matrix can be real or complex symmetric
   */
  template<class T, class Allocator>
  void GetInverse(Matrix<T, Symmetric, RowSymPacked, Allocator>& A, LapackInfo& info)
  {
    // factorisation of matrix A
    IVect IPivot;
    GetLU(A, IPivot);
    
    int n = A.GetM();
    T* ap = A.GetData();
    T zero, one, temp1, temp2, prod_scal;
    T valT, valD, ak, akp1, akkp1;
    SetComplexOne(one);
    SetComplexZero(zero);
    
    // Compute inv(A) from the factorization A = L*D*L'.
    
    // K is the main loop index, increasing from 1 to N in steps of
    // 1 or 2, depending on the size of the diagonal blocks.
    int k = n-1;
    int npp = A.GetDataSize();
    int kc = npp - 1;
    int kstep = 0;
    Vector<T> work(n);
    work.Fill(0);
    while (k >= 0)
      {
        int kcnext = kc - (n-k+1);
        
        if (IPivot(k) > 0)
          {
            // 1 x 1 diagonal block
            // invert the diagonal block
            ap[kc] = one / ap[kc];
            
            // Compute column K of the inverse.
            if (k < n-1)
              {
                for (int i = 1; i < n-k; i++)
                  {
                    work(i-1) = ap[kc+i];
                    ap[kc+i] = zero;
                  }
                
                int kk = kc+n-1-k+1;
                prod_scal = zero;
                for (int j = 0; j < n-1-k; j++)
                  {
                    temp1 = -work(j);
                    temp2 = zero;
                    ap[kc+j+1] += temp1*ap[kk];
                    int k2 = kk + 1;
                    for (int i = j+1; i < n-1-k; i++)
                      {
                        ap[kc+i+1] += temp1*ap[k2];
                        temp2 += ap[k2]*work(i);
                        k2++;
                      }
                    
                    ap[kc+j+1] -= temp2;
                    prod_scal += ap[kc+j+1]*work(j);
                    kk += n-1-k-j;
                  }
                
                ap[kc] -= prod_scal;
              }
            
            kstep = 1;
          }
        else
          {

            // 2 x 2 diagonal block
            
            // Invert the diagonal block.
            valT = ap[kcnext+1];
            ak = ap[kcnext] / valT;
            akp1 = ap[kc] / valT;
            akkp1 = ap[kcnext+1] / valT;
            valD = valT*( ak*akp1-one );
            ap[kcnext] = akp1 / valD;
            ap[kc] = ak / valD;
            ap[kcnext+1] = -akkp1 / valD;
            
            // Compute columns K-1 and K of the inverse.
            if (k < n-1)
              {
                for (int i = 1; i < n-k; i++)
                  {
                    work(i-1) = ap[kc+i];
                    ap[kc+i] = zero;
                  }
                
                int kk = kc+n-1-k+1;
                prod_scal = zero;
                for (int j = 0; j < n-1-k; j++)
                  {
                    temp1 = -work(j);
                    temp2 = zero;
                    ap[kc+j+1] += temp1*ap[kk];
                    int k2 = kk + 1;
                    for (int i = j+1; i < n-1-k; i++)
                      {
                        ap[kc+i+1] += temp1*ap[k2];
                        temp2 += ap[k2]*work(i);
                        k2++;
                      }
                    
                    ap[kc+j+1] -= temp2;
                    prod_scal += ap[kc+j+1]*work(j);
                    kk += n-1-k-j;
                  }
                
                ap[kc] -= prod_scal;
                
                prod_scal = zero;
                for (int j = 0; j < n-1-k; j++)
                  prod_scal += ap[kc+1+j]*ap[kcnext+2+j];
                
                ap[kcnext+1] -= prod_scal;
                for (int i = 1; i < n-k; i++)
                  {
                    work(i-1) = ap[kcnext+1+i];
                    ap[kcnext+1+i] = zero;
                  }
                
                kk = kc+n-1-k+1;
                prod_scal = zero;
                for (int j = 0; j < n-1-k; j++)
                  {
                    temp1 = -work(j);
                    temp2 = zero;
                    ap[kcnext+j+2] += temp1*ap[kk];
                    int k2 = kk + 1;
                    for (int i = j+1; i < n-1-k; i++)
                      {
                        ap[kcnext+i+2] += temp1*ap[k2];
                        temp2 += ap[k2]*work(i);
                        k2++;
                      }
                    
                    ap[kcnext+j+2] -= temp2;
                    prod_scal += ap[kcnext+j+2]*work(j);
                    kk += n-1-k-j;
                  }
                
                ap[kcnext] -= prod_scal;
              }
            
            kstep = 2;
            kcnext -= n-1-k+3;
          }
        
        int kp = abs(IPivot(k)) - 1;
        if (kp != k)
          {
            // Interchange rows and columns K and KP in the trailing
            // submatrix A(k-1:n,k-1:n)
            
            int kpc = npp - (n-1-kp+1)*(n-1-kp+2) / 2;
            if (kp < n-1)
              for (int i = 0; i < n-1-kp; i++)
                {
                  temp1 = ap[kc+kp-k+i+1];
                  ap[kc+kp-k+i+1] = ap[kpc+i+1];
                  ap[kpc+i+1] = temp1;
                }
            
            int kx = kc + kp - k;
            for (int j = k+1; j <= kp-1; j++)
              {
                kx += n-1-j+1;
                temp1 = ap[kc+j-k];
                ap[kc+j-k] = ap[kx];
                ap[kx] = temp1;
              }
            
            temp1 = ap[kc];
            ap[kc] = ap[kpc];
            ap[kpc] = temp1;
            
            if (kstep == 2)
              {
                temp1 = ap[kc-(n-1)+k-1];
                ap[kc-(n-1)+k-1] = ap[kc-(n-1)+kp-1];
                ap[kc-(n-1)+kp-1] = temp1;
              }

          }
        
        k -= kstep;
        kc = kcnext;
      }
  }
  
  
  //! A is replaced by its inverse
  /*!
    \param[inout] A on input, original matrix; on output, inverse
    The matrix can be real or complex unsymmetric
   */
  template<class T, class Storage, class Allocator>
  void GetInverse(Matrix<T, General, Storage, Allocator>& A, LapackInfo& info)
  {
    // factorisation of matrix A
    IVect IPivot;
    GetLU(A, IPivot);
    
    // forming inverse of U
    int n = A.GetM();
    T one, zero, temp;
    SetComplexOne(one);
    SetComplexZero(zero);
    for (int i = 0; i < n; i++)
      {
        for (int k = i+1; k < n; k++)
          {
            temp = zero;
            for (int j = i; j < k; j++)
              temp -= A(i, j)*A(j, k);
            
            A(i, k) = temp*A(k, k);
          }        
      }
    
    // solving equation inv(A)*L = inv(U) for inv(A)
    Vector<T> work(n);
    work.Fill(0);
    for (int k = n-2; k >= 0; k--)
      {
        for (int i = k+1; i < n; i++)
          {
            work(i) = A(i, k);
            A(i, k) = zero;
          }
        
        // Compute current column of inv(A).
        for (int i = 0; i < n; i++)
          for (int j = k+1; j < n; j++)
            A(i, k) -= A(i, j)*work(j);
      }

    // Apply column interchanges.
    for (int j = n-2; j >= 0; j--)
      {
        int jp = IPivot(j)-1;
        if (jp != j)
          {
            for (int k = 0; k < n; k++)
              {
                temp = A(k, j);
                A(k, j) = A(k, jp);
                A(k, jp) = temp;
              }
          }
      }
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, RowUpTriang, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }

  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, RowUpTriangPacked, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, ColUpTriang, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, ColUpTriangPacked, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, RowLoTriang, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, RowLoTriangPacked, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, ColLoTriang, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }


  template<class T, class Allocator>
  void GetInverse(Matrix<T, General, ColLoTriangPacked, Allocator>& A, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }
  
  
  //! replaces matrix A by its Cholesky factorisation
  template<class T, class Prop, class Storage, class Allocator>
  void GetCholesky(Matrix<T, Prop, Storage, Allocator>& A, LapackInfo& info)
  {
    int n = A.GetM();
    T val, invVal, vloc, one, zero;
    SetComplexOne(one);
    SetComplexZero(zero);
    for (int j = 0; j < n; j++)
      {
	// diagonal element
	val = A(j, j);
	for (int k = 0; k < j; k++)
	  val -= A(k, j)*A(k, j);
	
        if (val <= zero)
	  {
            cout << "Error during Cholesky factorization " << endl;
            cout << "Matrix must be definite positive " << endl;
            cout << "but diagonal element of row " << j 
                 << " is equal to " << val << endl;

	    throw WrongArgument("GetCholesky", "Negative diagonal");
          }
	
	val = sqrt(val);
	A.Get(j, j) = val;
	invVal = one/val;
	
	// other elements of the row
	for (int k = j+1; k < n; k++)
	  {
	    vloc = A(j, k);
	    for (int m = 0; m < j; m++)
	      vloc -= A(k, m)*A(m, j);
	    
	    A.Get(j, k) = vloc*invVal;
	  }
      }
  }


  //! replaces matrix A by its Cholesky factorisation (Hermitian matrix)
  template<class T, class Prop, class Storage, class Allocator>
  void GetCholesky(Matrix<complex<T>, Prop, Storage, Allocator>& A, LapackInfo& info)
  {
    int n = A.GetM();
    T val, invVal, one, zero;
    SetComplexOne(one);
    SetComplexZero(zero);
    complex<T> vloc;
    for (int j = 0; j < n; j++)
      {
	// diagonal element
	val = real(A(j, j));
	for (int k = 0; k < j; k++)
	  val -= absSquare(A(k, j));
	
        if (val <= zero)
	  {
            cout << "Error during Cholesky factorization " << endl;
            cout << "Matrix must be definite positive " << endl;
            cout << "but diagonal element of row " << j 
                 << " is equal to " << val << endl;

	    throw WrongArgument("GetCholesky", "Negative diagonal");
          }
	
	val = sqrt(val);
	A.Get(j, j) = complex<T>(val, zero);
	invVal = one/val;
	
	// other elements of the row
	for (int k = j+1; k < n; k++)
	  {
	    vloc = A(j, k);
	    for (int m = 0; m < j; m++)
              vloc -= A(m, k)*conjugate(A(m, j));
              	    
	    A.Get(j, k) = vloc*invVal;
	  }
      }
  }
  
  
  //! when A = L L^T, solves L x = b or L^T x = b
  /*!
    \param[in] TransA SeldonNoTrans for solving L x = b, SeldonTrans for L^T x = b
    \param[in] A Cholesky factorisation of matrix A
    \param[inout] x on input, right hand side b; on output, solution x
    We assume here that GetCholesky has been called before to generate
    the Cholesky factorisation of the original matrix
   */
  template<class T, class Prop,
           class Storage, class Allocator,
	   class T2, class Storage2, class Allocator2>
  void SolveCholesky(const SeldonTranspose& TransA,
		     const Matrix<T, Prop, Storage, Allocator>& A,
		     Vector<T2, Storage2, Allocator2>& x, LapackInfo& info)
  {
    int n = A.GetM();
    if (n <= 0)
      return;
    
    if (TransA.Trans())
      {
        // we solve L^T x = x
	for (int i = n-1; i >= 0; i--) 
          {
            T2 val = x(i);
            for (int k = i+1; k < n; k++)
	      val -= A(i, k)*x(k);
                          
            x(i) = val/A(i, i);
          }
      }
    else
      {
        // we solve L x = x
        for (int i = 0; i < n; i++) 
          {
            x(i) /= A(i, i);
            for (int k = i+1; k < n; k++)
	      x(k) -= A(i, k)*x(i);
	    
          }
      }
  }
  
  
  //! when A = L L^T, computes L x or L^T x
  /*!
    \param[in] TransA SeldonNoTrans for computing L x, SeldonTrans for L^T x
    \param[in] A Cholesky factorisation of matrix A
    \param[inout] x x is overwritten by L x or L^T x
    We assume here that GetCholesky has been called before to generate
    the Cholesky factorisation of the original matrix
   */
  template<class T,
           class Prop, class Storage, class Allocator,
	   class T2, class Storage2, class Allocator2>
  void MltCholesky(const SeldonTranspose& TransA,
                   const Matrix<T, Prop, Storage, Allocator>& A,
                   Vector<T2, Storage2, Allocator2>& x, LapackInfo& info)
  {
    int n = A.GetM();
    if (n <= 0)
      return;
    
    if (TransA.Trans())
      {
        // we overwrite x by L^T x
	for (int i = 0; i < n; i++) 
          {
            T2 val = x(i)*A(i, i);
            for (int k = i+1; k < n ; k++)
	      val += A(i, k)*x(k);
	    
            x(i) = val;
          }
      }
    else
      {
        // we overwrite x by L x
        for (int i = n-1; i >= 0; i--) 
          {
            for (int k = i+1; k < n; k++)
	      x(k) += A(i, k)*x(i);
            
            x(i) *= A(i, i);
          }
      }
  }


  //! when A = L L^T, solves L x = b or L^T x = b
  /*!
    \param[in] TransA SeldonNoTrans for solving L x = b, SeldonTrans for L^T x = b
    \param[in] A Cholesky factorisation of matrix A
    \param[inout] x on input, right hand side b; on output, solution x
    We assume here that GetCholesky has been called before to generate
    the Cholesky factorisation of the original matrix
   */
  template<class T, class Prop,
           class Storage, class Allocator,
	   class T2, class Storage2, class Allocator2>
  void SolveCholesky(const SeldonTranspose& TransA,
		     const Matrix<complex<T>, Prop, Storage, Allocator>& A,
		     Vector<complex<T2>, Storage2, Allocator2>& x, LapackInfo& info)
  {
    int n = A.GetM();
    if (n <= 0)
      return;
    
    if (TransA.Trans())
      {
        // we solve L^T x = x
	for (int i = n-1; i >= 0; i--) 
          {
            complex<T> val = x(i);
            for (int k = i+1; k < n; k++)
	      val -= conjugate(A(i, k))*x(k);
                          
            x(i) = val/A(i, i);
          }
      }
    else if (TransA.ConjTrans())
      {
        // we solve L^T x = x
	for (int i = n-1; i >= 0; i--) 
          {
            complex<T> val = x(i);
            for (int k = i+1; k < n; k++)
	      val -= A(i, k)*x(k);
                          
            x(i) = val/A(i, i);
          }
      }
    else
      {
        // we solve L x = x
        for (int i = 0; i < n; i++) 
          {
            x(i) /= A(i, i);
            for (int k = i+1; k < n; k++)
	      x(k) -= conjugate(A(i, k))*x(i);
	    
          }
      }
  }
  
  
  //! when A = L L^T, computes L x or L^T x
  /*!
    \param[in] TransA SeldonNoTrans for computing L x, SeldonTrans for L^T x
    \param[in] A Cholesky factorisation of matrix A
    \param[inout] x x is overwritten by L x or L^T x
    We assume here that GetCholesky has been called before to generate
    the Cholesky factorisation of the original matrix
   */
  template<class T,
           class Prop, class Storage, class Allocator,
	   class T2, class Storage2, class Allocator2>
  void MltCholesky(const SeldonTranspose& TransA,
                   const Matrix<complex<T>, Prop, Storage, Allocator>& A,
                   Vector<complex<T2>, Storage2, Allocator2>& x, LapackInfo& info)
  {
    int n = A.GetM();
    if (n <= 0)
      return;
    
    if (TransA.Trans())
      {
        // we overwrite x by L^T x
	for (int i = 0; i < n; i++) 
          {
            complex<T> val = x(i)*A(i, i);
            for (int k = i+1; k < n ; k++)
	      val += conjugate(A(i, k))*x(k);
	    
            x(i) = val;
          }
      }
    else if (TransA.ConjTrans())
      {
        // we overwrite x by L^T x
	for (int i = 0; i < n; i++) 
          {
            complex<T> val = x(i)*A(i, i);
            for (int k = i+1; k < n ; k++)
	      val += A(i, k)*x(k);
	    
            x(i) = val;
          }
      }
    else
      {
        // we overwrite x by L x
        for (int i = n-1; i >= 0; i--) 
          {
            for (int k = i+1; k < n; k++)
	      x(k) += conjugate(A(i, k))*x(i);
            
            x(i) *= A(i, i);
          }
      }
  }
  
  
  //! replaces matrix A by QR factorisation
  /*!
    \param[inout] A on input, original matrix; on output, QR factorisation
    \param[out] tau intermediary array needed to SolveQR
   */
  template<class T, class Prop, class Storage, class Allocator, class Allocator2>
  void GetQR(Matrix<T, Prop, Storage, Allocator>& A, Vector<T, VectFull, Allocator2>& tau,
	     LapackInfo& info)
  {
    T one, zero; SetComplexOne(one);
    SetComplexZero(zero);
    typedef typename ClassComplexType<T>::Treal Treal;
    Treal normU, normUp, normN, invNormN;
    T v1, prod_scal, alpha, n1;
    SetComplexZero(normU);
    int m = A.GetM();
    int n = A.GetN();
    tau.Reallocate(n); tau.Fill(1);
    for (int i = 0; i < n; i++)
      {
        // U = (a_ii, a_{i+1,i}, ..., a_{i, n})
        // V = (-signe(a_ii) || U ||, 0, 0, ..., 0)
        // N = (U-V)/ || U - V ||
        normUp = 0;
        for (int k = i+1; k < m; k++)
          normUp += absSquare(A(k, i));
        
        if (normUp > Treal(0))
          {
            // norm of U
            normU = sqrt(normUp + absSquare(A(i, i)));
	    SetComplexReal(normU, v1);
	    if (A(i, i) != zero)
	      v1 *= -A(i, i) / abs(A(i, i));
	    
            // norm of N and first component of N
            alpha = A(i, i) - v1; 
            normN = sqrt(normUp + absSquare(alpha));
            invNormN = Treal(1)/normN;
            n1 = alpha*invNormN;
                        
            // first-column is replaced with normale except first component
            A(i, i) = v1;
            for (int k = i+1; k < m; k++)
              A(k, i) *= invNormN;
            
            // we apply H = I - 2 N N^T to other sub-columns of A
            for (int j = i+1; j < n; j++)
              {
                // computation of C \cdot N
                prod_scal = A(i, j)*n1;
                for (int k = i+1; k < m; k++)
                  prod_scal += A(k, j)*A(k, i);
                
                prod_scal *= T(2);
                // we subtract - 2 N ( N^T cdot C )
                A(i, j) -= n1*prod_scal;
                for (int k = i+1; k < m; k++)
                  A(k, j) -= A(k, i)*prod_scal;
              }
            
            // first component of normale is stored in tau
            tau(i) = n1;
          }
      }
  }

  
  //! overwrites x by solution of Q R x = b
  /*!
    \param[in] A QR factorisation of a matrix
    \param[in] tau intermediary array produced by GetQR
    \param[inout] x on input, right hand side; on output, solution x
   */
  template<class T, class Prop, class Storage, class Allocator,
           class T2, class Allocator2, class Allocator3>
  void SolveQR(const Matrix<T, Prop, Storage, Allocator>& A,
               const Vector<T, VectFull, Allocator2>& tau,
               Vector<T2, VectFull, Allocator3>& X, LapackInfo& info)
  {
    T prod_scal, val, one;
    SetComplexOne(one);

    int m = A.GetM();
    int n = A.GetN();
    // computation of Q^T x
    for (int i = 0; i < n; i++)
      {
        // we overwrite X with X - 2 N N^t X
        if (tau(i) != one)
          {
            prod_scal = tau(i)*X(i);
            for (int j = i+1; j < m; j++)
              prod_scal += A(j, i)*X(j);
            
            prod_scal *= T(2);
            X(i) -= prod_scal*tau(i);
            for (int j = i+1; j < m; j++)
              X(j) -= prod_scal*A(j, i);
          }
      }

    // now resolution of R x = Q^T x
    X.Resize(n);
    for (int i = n-1; i >= 0; i--)
      {
        val = X(i);
        for (int j = i+1; j < n; j++)
          val -= A(i, j)*X(j);
        
        X(i) = val/A(i, i);        
      }
  }
  
  
  //! Multiplies X by Q from QR factorisation of A
  /*!
    \param[in] trans SeldonNoTrans or SeldonTrans
    \param[in] A QR factorisation of the original matrix A
    \param[in] tau intermediary array filled by GetQR
    \param[inout] X on input X, on output Q X or Q^T X
    Once QR factorisation of A (A = Q R) has been performed (by calling GetQR),
    this function overwrites X by Q X or Q^T X if trans is equal to class_SeldonTrans
   */
  template<class IsTranspose, class T, class Prop,
	   class Storage, class Allocator,
           class T2, class Allocator2, class Allocator3>
  void MltQ_FromQR(const IsTranspose& trans,
		   const Matrix<T, Prop, Storage, Allocator>& A,
		   const Vector<T, VectFull, Allocator2>& tau,
		   Vector<T2, VectFull, Allocator3>& X, LapackInfo& info)
  {
    if (!trans.Trans())
      {
	abort();
      }
    
    T prod_scal, val, one;
    SetComplexOne(one);
    int m = A.GetM();
    int n = A.GetN();
    // computation of Q^T x
    for (int i = 0; i < n; i++)
      {
        // we overwrite X with X - 2 N N^t X
        if (tau(i) != one)
          {
            prod_scal = tau(i)*X(i);
            for (int j = i+1; j < m; j++)
              prod_scal += A(j, i)*X(j);
            
            prod_scal *= T(2);
            X(i) -= prod_scal*tau(i);
            for (int j = i+1; j < m; j++)
              X(j) -= prod_scal*A(j, i);
          }
            
      }
  }
  
  
  //! overwrites x by solution of A y = x, when A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void SolveUpper(const Matrix<T, General, Storage, Allocator1>& A,
                  Vector<T, VectFull, Allocator2>& x)
  {
    T vloc;
    int n = x.GetM();
    for (int i = n-1; i >= 0; i--)
      {
        vloc = x(i);
        for (int j = i+1; j < n; j++)
          vloc -= A(i, j)*x(j);
        
        x(i) = vloc/A(i, i);
      }
  }


  //! overwrites x by solution of A y = x or A^T y = x or A^H y = x, when A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void SolveUpper(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
                  const Matrix<T, General, Storage, Allocator1>& A,
                  Vector<T, VectFull, Allocator2>& x)
  {    
    T vloc;
    int n = x.GetM();
    if (DiagA.Unit())
      {
        if (TransA.Trans())
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = 0; j < i; j++)
                vloc -= A(j, i)*x(j);
              
              x(i) = vloc;
            }
        else if (TransA.ConjTrans())
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = 0; j < i; j++)
                vloc -= conjugate(A(j, i))*x(j);
              
              x(i) = vloc;
            }
        else
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc -= A(i, j)*x(j);
              
              x(i) = vloc;
            }
      }
    else
      {
        if (TransA.Trans())
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = 0; j < i; j++)
                vloc -= A(j, i)*x(j);
              
              x(i) = vloc/A(i, i);
            }
        else if (TransA.ConjTrans())
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = 0; j < i; j++)
                vloc -= conjugate(A(j, i))*x(j);
              
              x(i) = vloc/conjugate(A(i, i));
            }
        else
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc -= A(i, j)*x(j);
              
              x(i) = vloc/A(i, i);
            }
      }
  }

  
  //! overwrites x by solution of A y = x, when A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void SolveLower(const Matrix<T, General, Storage, Allocator1>& A,
                  Vector<T, VectFull, Allocator2>& x)
  {
    T vloc;
    int n = x.GetM();
    for (int i = 0; i < n; i++)
      {
        vloc = x(i);
        for (int j = 0; j < i; j++)
          vloc -= A(i, j)*x(j);
        
        x(i) = vloc/A(i, i);
      }
  }


  //! overwrites x by solution of A y = x or A^T y = x or A^H y = x, when A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void SolveLower(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
                  const Matrix<T, General, Storage, Allocator1>& A,
                  Vector<T, VectFull, Allocator2>& x)
  {    
    T vloc;
    int n = x.GetM();
    if (DiagA.Unit())
      {
        if (TransA.Trans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc -= A(j, i)*x(j);
              
              x(i) = vloc;
            }
        else if (TransA.ConjTrans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc -= conjugate(A(j, i))*x(j);
              
              x(i) = vloc;
            }
        else
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = 0; j < i; j++)
                vloc -= A(i, j)*x(j);
              
              x(i) = vloc;
            }
      }
    else
      {
        if (TransA.Trans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc -= A(j, i)*x(j);
              
              x(i) = vloc/A(i, i);
            }
        else if (TransA.ConjTrans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc -= conjugate(A(j, i))*x(j);
              
              x(i) = vloc/conjugate(A(i, i));
            }
        else
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = 0; j < i; j++)
                vloc -= A(i, j)*x(j);
              
              x(i) = vloc/A(i, i);
            }
      }
  }

  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, ColUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, ColUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, RowUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, RowUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveUpper(TransA, DiagA, A, x);
  }
  

  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, ColLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, ColLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, RowLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Solve(const Matrix<T, General, RowLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Solve(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    SolveLower(TransA, DiagA, A, x);
  }

  
  //! replaces B by alpha A^{-1} B or alpha B A^{-1}
  template<class T, class Prop, class Storage, class Allocator,
           class Prop0, class Storage0, class Allocator0>
  void Solve(const SeldonSide& side, const T& alpha,
             const Matrix<T, Prop, Storage, Allocator>& A,
             Matrix<T, Prop0, Storage0, Allocator0>& B)
  {
    Vector<T> x;
    if (side.Left())
      for (int i = 0; i < B.GetN(); i++)
        {
          GetCol(B, i, x);
          Solve(A, x);
          Mlt(alpha, x);
          SetCol(x, i, B);
        }
    else
      {
        for (int i = 0; i < B.GetM(); i++)
          {
            GetRow(B, i, x);
            Solve(SeldonTrans, SeldonNonUnit, A, x);
            Mlt(alpha, x);
            SetRow(x, i, B);
          }
      }
  }


  //! replaces B by alpha A^{-1} B or alpha B A^{-1}
  template<class T, class Prop, class Storage, class Allocator,
           class Prop0, class Storage0, class Allocator0>
  void Solve(const SeldonSide& side, const T& alpha,
             const SeldonTranspose& TransA,
             const SeldonDiag& DiagA,
             const Matrix<T, Prop, Storage, Allocator>& A,
             Matrix<T, Prop0, Storage0, Allocator0>& B)
  {
    Vector<T> x;
    if (side.Left())
      for (int i = 0; i < B.GetN(); i++)
        {
          GetCol(B, i, x);
          Solve(TransA, DiagA, A, x);
          Mlt(alpha, x);
          SetCol(x, i, B);
        }
    else
      {
        if (TransA.Trans())
          for (int i = 0; i < B.GetM(); i++)
            {
              GetRow(B, i, x);
              Solve(SeldonNoTrans, DiagA, A, x);
              Mlt(alpha, x);
              SetRow(x, i, B);
            }
        else if (TransA.ConjTrans())
          for (int i = 0; i < B.GetM(); i++)
            {
              GetRow(B, i, x); Conjugate(x);
              Solve(SeldonNoTrans, DiagA, A, x);
              Conjugate(x); Mlt(alpha, x);
              SetRow(x, i, B);
            }
        else
          for (int i = 0; i < B.GetM(); i++)
            {
              GetRow(B, i, x);
              Solve(SeldonTrans, DiagA, A, x);
              Mlt(alpha, x);
              SetRow(x, i, B);
            }
      }
  }
  
  
  //! overwrites x by A x where A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void MltUpper(const Matrix<T, General, Storage, Allocator1>& A,
                  Vector<T, VectFull, Allocator2>& x)
  {
    T vloc;
    int n = x.GetM();
    for (int i = 0; i < n; i++)
      {
        vloc = A(i, i)*x(i);
        for (int j = i+1; j < n; j++)
          vloc += A(i, j)*x(j);
        
        x(i) = vloc;
      }
  }


  //! overwrites x by A x or A^T x or A^H x where A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void MltUpper(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
                const Matrix<T, General, Storage, Allocator1>& A,
                Vector<T, VectFull, Allocator2>& x)
  {    
    T vloc;
    int n = x.GetM();
    if (DiagA.Unit())
      {
        if (TransA.Trans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i-1; j >= 0; j--)
                vloc += A(j, i)*x(j);
              
              x(i) = vloc;
            }
        else if (TransA.ConjTrans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i-1; j >= 0; j--)
                vloc += conjugate(A(j, i))*x(j);
              
              x(i) = vloc;
            }
        else
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc += A(i, j)*x(j);
              
              x(i) = vloc;
            }
      }
    else
      {
        if (TransA.Trans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = A(i, i)*x(i);
              for (int j = i-1; j >= 0; j--)
                vloc += A(j, i)*x(j);
              
              x(i) = vloc;
            }
        else if (TransA.ConjTrans())
          for (int i = n-1; i >= 0; i--)
            {
              vloc = conjugate(A(i, i))*x(i);
              for (int j = i-1; j >= 0; j--)
                vloc += conjugate(A(j, i))*x(j);
              
              x(i) = vloc;
            }
        else
          for (int i = 0; i < n; i++)
            {
              vloc = A(i, i)*x(i);
              for (int j = i+1; j < n; j++)
                vloc += A(i, j)*x(j);
              
              x(i) = vloc;
            }
      }
  }

  
  //! overwrites x by A x or A^T x or A^H x where A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void MltLower(const Matrix<T, General, Storage, Allocator1>& A,
                  Vector<T, VectFull, Allocator2>& x)
  {
    T vloc;
    int n = x.GetM();
    for (int i = n-1; i >= 0; i--)
      {
        vloc = A(i, i)*x(i);
        for (int j = i-1; j >= 0; j--)
          vloc += A(i, j)*x(j);
        
        x(i) = vloc;
      }
  }


  //! overwrites x by A x or A^T or A^H x where A is triangular
  template<class T, class Allocator1, class Storage, class Allocator2>
  void MltLower(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
                const Matrix<T, General, Storage, Allocator1>& A,
                Vector<T, VectFull, Allocator2>& x)
  {    
    T vloc;
    int n = x.GetM();
    if (DiagA.Unit())
      {
        if (TransA.Trans())
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc += A(j, i)*x(j);
              
              x(i) = vloc;
            }
        else if (TransA.ConjTrans())
          for (int i = 0; i < n; i++)
            {
              vloc = x(i);
              for (int j = i+1; j < n; j++)
                vloc += conjugate(A(j, i))*x(j);
              
              x(i) = vloc;
            }
        else
          for (int i = n-1; i >= 0; i--)
            {
              vloc = x(i);
              for (int j = i-1; j >= 0; j--)
                vloc += A(i, j)*x(j);
              
              x(i) = vloc;
            }
      }
    else
      {
        if (TransA.Trans())
          for (int i = 0; i < n; i++)
            {
              vloc = A(i, i)*x(i);
              for (int j = i+1; j < n; j++)
                vloc += A(j, i)*x(j);
              
              x(i) = vloc;
            }
        else if (TransA.ConjTrans())
          for (int i = 0; i < n; i++)
            {
              vloc = conjugate(A(i, i))*x(i);
              for (int j = i+1; j < n; j++)
                vloc += conjugate(A(j, i))*x(j);
              
              x(i) = vloc;
            }
        else
          for (int i = n-1; i >= 0; i--)
            {
              vloc = A(i, i)*x(i);
              for (int j = i-1; j >= 0; j--)
                vloc += A(i, j)*x(j);
              
              x(i) = vloc;
            }
      }
  }

  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, ColUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, ColUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, RowUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowUpTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, RowUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowUpTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltUpper(TransA, DiagA, A, x);
  }
  

  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, ColLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, ColLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, ColLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, RowLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowLoTriang, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(TransA, DiagA, A, x);
  }


  template<class T, class Allocator1, class Allocator2>
  void Mlt(const Matrix<T, General, RowLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(A, x);
  }
  
  
  template<class T, class Allocator1, class Allocator2>
  void Mlt(const SeldonTranspose& TransA, const SeldonDiag& DiagA,
             const Matrix<T, General, RowLoTriangPacked, Allocator1>& A,
             Vector<T, VectFull, Allocator2>& x)
  {
    MltLower(TransA, DiagA, A, x);
  }


  //! replaces B by alpha A B or alpha B A
  template<class T, class Prop, class Storage, class Allocator,
           class Prop0, class Storage0, class Allocator0>
  void Mlt(const SeldonSide& side, const T& alpha,
           const Matrix<T, Prop, Storage, Allocator>& A,
           Matrix<T, Prop0, Storage0, Allocator0>& B)
  {
    Vector<T> x;
    if (side.Left())
      for (int i = 0; i < B.GetN(); i++)
        {
          GetCol(B, i, x);
          Mlt(A, x);
          Mlt(alpha, x);
          SetCol(x, i, B);
        }
    else
      {
        for (int i = 0; i < B.GetM(); i++)
          {
            GetRow(B, i, x);
            Mlt(SeldonTrans, SeldonNonUnit, A, x);
            Mlt(alpha, x);
            SetRow(x, i, B);
          }
      }
  }


  //! replaces B by alpha A B or alpha B A
  template<class T, class Prop, class Storage, class Allocator,
           class Prop0, class Storage0, class Allocator0>
  void Mlt(const SeldonSide& side, const T& alpha,
           const SeldonTranspose& TransA,
           const SeldonDiag& DiagA,
           const Matrix<T, Prop, Storage, Allocator>& A,
           Matrix<T, Prop0, Storage0, Allocator0>& B)
  {
    Vector<T> x;
    if (side.Left())
      for (int i = 0; i < B.GetN(); i++)
        {
          GetCol(B, i, x);
          Mlt(TransA, DiagA, A, x);
          Mlt(alpha, x);
          SetCol(x, i, B);
        }
    else
      {
        if (TransA.Trans())
          for (int i = 0; i < B.GetM(); i++)
            {
              GetRow(B, i, x);
              Mlt(SeldonNoTrans, DiagA, A, x);
              Mlt(alpha, x);
              SetRow(x, i, B);
            }
        else if (TransA.ConjTrans())
          for (int i = 0; i < B.GetM(); i++)
            {
              GetRow(B, i, x); Conjugate(x);
              Mlt(SeldonNoTrans, DiagA, A, x);
              Conjugate(x); Mlt(alpha, x);
              SetRow(x, i, B);
            }
        else
          for (int i = 0; i < B.GetM(); i++)
            {
              GetRow(B, i, x);
              Mlt(SeldonTrans, DiagA, A, x);
              Mlt(alpha, x);
              SetRow(x, i, B);
            }
      }
  }

  
  //! Adds alpha X. Y^T to matrix A
  template<class T, class Alloc1, class Alloc2, class Storage, class Alloc0>
  void Rank1Update(const T& alpha, const Vector<T, VectFull, Alloc1>& X,
                   const Vector<T, VectFull, Alloc2>& Y,
                   Matrix<T, General, Storage, Alloc0>& A)
  {
    for (int i = 0; i < A.GetM(); i++)
      for (int j = 0; j < A.GetN(); j++)
        A(i, j) += alpha*X(i)*Y(j);
  }

  //! Adds alpha X. Y^H or alpha X.Y^T to matrix A
  template<class T, class Alloc1, class Alloc2, class Storage, class Alloc0>
  void Rank1Update(const T& alpha, const Vector<T, VectFull, Alloc1>& X,
                   const SeldonConjugate& ConjY, const Vector<T, VectFull, Alloc2>& Y,
                   Matrix<T, General, Storage, Alloc0>& A)
  {
    if (ConjY.Conj())
      for (int i = 0; i < A.GetM(); i++)
        for (int j = 0; j < A.GetN(); j++)
          A(i, j) += alpha*X(i)*conjugate(Y(j));
    else
      for (int i = 0; i < A.GetM(); i++)
        for (int j = 0; i < A.GetN(); j++)
          A(i, j) += alpha*X(i)*Y(j);
  }

  
  //! Adds alpha X.X^T to matrix A
  template<class T, class Alloc1, class Storage, class Alloc0>
  void Rank1Update(const T& alpha, const Vector<T, VectFull, Alloc1>& X,
                   Matrix<T, Symmetric, Storage, Alloc0>& A)
  {
    for (int i = 0; i < A.GetM(); i++)
      for (int j = i; j < A.GetM(); j++)
        A.Get(i, j) += alpha*X(i)*X(j);
  }


  //! Adds alpha X.X^H to matrix A
  template<class T, class Alloc1, class Storage, class Alloc0>
  void Rank1Update(const T& alpha, const Vector<complex<T>, VectFull, Alloc1>& X,
                   Matrix<complex<T>, Hermitian, Storage, Alloc0>& A)
  {
    for (int i = 0; i < A.GetM(); i++)
      for (int j = i; j < A.GetM(); j++)
        A.Get(i, j) += alpha*X(i)*conjugate(X(j));
  }


  //! Adds alpha X.Y^T + Y.X^T to matrix A
  template<class T, class Alloc1, class Alloc2, class Storage, class Alloc0>
  void Rank2Update(const T& alpha, const Vector<T, VectFull, Alloc1>& X,
                   const Vector<T, VectFull, Alloc2>& Y,
                   Matrix<T, Symmetric, Storage, Alloc0>& A)
  {
    for (int i = 0; i < A.GetM(); i++)
      for (int j = i; j < A.GetM(); j++)
        A.Get(i, j) += alpha*(X(i)*Y(j) + X(j)*Y(i));
  }


  //! Adds alpha X.Y^H + Y.X^H to matrix A
  template<class T, class Alloc1, class Alloc2, class Storage, class Alloc0>
  void Rank2Update(const T& alpha, const Vector<complex<T>, VectFull, Alloc1>& X,
                   const Vector<complex<T>, VectFull, Alloc2>& Y,
                   Matrix<complex<T>, Hermitian, Storage, Alloc0>& A)
  {
    for (int i = 0; i < A.GetM(); i++)
      for (int j = i; j < A.GetM(); j++)
        A.Get(i, j) += alpha*(X(i)*conjugate(Y(j)) + conjugate(X(j))*Y(i));
  }
  
  
  //! computes C = beta C + alpha A B or beta C + alpha B A depending on the side
  template<class T0, class T1, class Prop1, class Storage1, class Allocator1,
           class T2, class Prop2, class Storage2, class Allocator2,
           class T3, class T4, class Prop4, class Storage4, class Allocator4>
  void MltAdd(const SeldonSide& Side, const T0& alpha,
              const Matrix<T1, Prop1, Storage1, Allocator1>& A,
              const Matrix<T2, Prop2, Storage2, Allocator2>& B,
              const T3& beta, Matrix<T4, Prop4, Storage4, Allocator4>& C)
  {
    if (Side.Left())
      MltAdd(alpha, A, B, beta, C);
    else
      MltAdd(alpha, B, A, beta, C);
  }
#else
  template<class T, class Prop, class Allocator1,
           class Storage2, class Allocator2>
  void SolveCholesky(const SeldonTranspose& TransA,
		     const Matrix<T, Prop, RowSymPacked, Allocator1>& A,
		     Vector<complex<T>, Storage2, Allocator2>& x, LapackInfo& info)
  {
    int N = x.GetM();
    Vector<T> xr(N), xi(N);
    for (int i = 0; i < N; i++)
      {
	xr(i) = real(x(i));
	xi(i) = imag(x(i));
      }
    
    SolveCholesky(TransA, A, xr);
    SolveCholesky(TransA, A, xi);

    for (int i = 0; i < N; i++)
      x(i) = complex<T>(xr(i), xi(i));
    
  }
#endif


  template<class T, class Prop, class Storage, class Allocator,
           class Allocator2, class Allocator3>
  void SolveQR(const Matrix<T, Prop, Storage, Allocator>& A,
               const Vector<T, VectFull, Allocator2>& tau,
               Vector<complex<T>, VectFull, Allocator3>& x, LapackInfo& info)
  {
    cout << "Not implemented" << endl;
    abort();
  }

  
  //! Computes Householder transformation for a single vector
  /*!
    The vector U is replaced by the normale defining the Householder transformation
    H = I - 2 N N^T
    The function returns the first component of vector V
   */
  template<class T>
  T GetHouseholderNormale(Vector<T>& U)
  {
    if (U.GetM() <= 0)
      return T(0);
    
    // the same kind of computation is made in GetQR
    
    // we start from a vector U:
    // U = (U_0, U_{1}, ..., U_{n-1})

    // and we want to find an orthogonal transformation into a vector V:
    // V = (-sign(U_0) || U ||, 0, 0, ..., 0)

    // sign(U_0) is put to maximize the first component of U-V

    // The transformation to transform U into V is the Householder transformation
    // H = I - 2 N N^T  where N is the normale defined as
    // N = (U-V)/ || U - V ||
    typedef typename ClassComplexType<T>::Treal Treal;
    Treal normUp = 0;
    for (int k = 1; k < U.GetM(); k++)
      normUp += absSquare(U(k));
    
    T zero; SetComplexZero(zero);
    T v1 = U(0);
    if (normUp > Treal(0))
      {
	// norm of U
	Treal normU = sqrt(normUp + absSquare(U(0)));
	// first component of V
	SetComplexReal(normU, v1);
	if (U(0) != zero)
	  v1 *= -U(0) / abs(U(0));
	
	// norm of N and first component of N
	T alpha = U(0) - v1; 
	Treal normN = sqrt(normUp + absSquare(alpha));
	Treal invNormN = Treal(1)/normN;
	T n1 = alpha*invNormN;
	
	// U is replaced by N
	U(0) = n1;
	for (int k = 1; k < U.GetM(); k++)
	  U(k) *= invNormN;	
      }
    else
      {
	U.Zero();
	SetComplexOne(U(0));
      }
    
    // we return the first component of V
    return v1;
  }

  
  //! Q is orthogonalized with respect to householder transformations stored in H
  /*!
    Householder transformation for Q(size:end) is computed and stored in H(size)
    Householder transformations are applied to the canonical vector [0, .., 1, 0, .., 0]
    such that Q will be orthogonalized with respect to Householder transformations stored in H
   */
  template<class T>
  void GetReorthogonalization(Vector<Vector<T> >& H, int size, Vector<T>& Q)
  {
    T one; SetComplexOne(one);
    int dim = Q.GetM() - size;
    H(size).Reallocate(dim);
    if (dim > 1)
      {
	for (int i = 0; i < dim; i++)
	  H(size)(i) = Q(i+size);
	
	GetHouseholderNormale(H(size));
      }
    else
      H(size) = one;

    Q.Zero();
    Q(size) = one;
    for (int i = size; i >= 0; i--)
      ApplyHouseholderTransformation(H(i), i, Q);
  }
  

  //! Applies Householder transformation to a vector
  /*!
    Householder transformation is given as H = I - 2 N N^H
    Only the normale N is provided as input array
    The transformation is applied to X(k:end)
   */
  template<class T>
  void ApplyHouseholderTransformation(const Vector<T>& N, int k, Vector<T>& X)
  {
    // scalar product with N
    T x_scal_N;
    SetComplexZero(x_scal_N);
    for (int i = k; i < X.GetM(); i++)
      x_scal_N += conjugate(N(i-k))*X(i);

    // replacing X by (X - 2 (N \cdot X) X)
    x_scal_N *= 2;
    for (int i = k; i < X.GetM(); i++)
      X(i) -= x_scal_N*N(i-k);    
    
  }
  
  
  //! Replaces A by its pseudo-inverse
  template<class T0, class T1, class Prop1, class Storage1, class Allocator1>
  void GetPseudoInverse(Matrix<T0, Prop1, Storage1, Allocator1>& A, const T1& epsilon,
			LapackInfo& info)
  {
    // not implemented in the general case
    cout << "Not implemented for any matrix" << endl;
    abort(); 
  }
  
}

#define SELDON_FILE_FACTORIZATION_LU_CXX
#endif
