#ifndef MONTJOIE_FILE_FINITE_ELEMENT_HCURL_AXI_CXX

namespace Montjoie
{
  
  ElementReference<Dimension2, 4>
  ::ElementReference(ElementReference<Dimension2, 1>& Fb_h1, ElementReference<Dimension2, 2>& Fb_hcurl,
		     ElementGeomReference<Dimension2>& Fb_geom)
    : ElementGeomReferenceContainer<Dimension2>(Fb_geom), fe_h1(Fb_h1), fe_hcurl(Fb_hcurl)
  {
    Fb_geom.SetAxisymmetricGeometry();
  }
  
  
  size_t ElementReference<Dimension2, 4>::GetMemorySize() const
  {
    size_t taille = fe_h1.GetMemorySize() + fe_hcurl.GetMemorySize();
    taille += ScalarDof.GetMemorySize() + VectorialDof.GetMemorySize();
    taille += ScalarPtsDof.GetMemorySize() + VectorialPtsDof.GetMemorySize();
    taille += ScalarQuadrature.GetMemorySize() + VectorialQuadrature.GetMemorySize();
    taille += points2d_inside.GetMemorySize();
    taille += EdgesDof.GetMemorySize() + LocalScalNumber.GetMemorySize() + LocalVecNumber.GetMemorySize();
    taille += Seldon::GetMemorySize(ScalarPtsDofFace) + Seldon::GetMemorySize(VectorialPtsDofFace);
    taille += Seldon::GetMemorySize(num_quad_points_surf);
    taille += power_two_face.GetMemorySize() +  PosDofOnFace.GetMemorySize();
    taille += ChE.GetMemorySize() + ChH.GetMemorySize() + ChEteta.GetMemorySize()
      + ChHteta.GetMemorySize() + RhE.GetMemorySize() + RhEteta.GetMemorySize();
    taille += RhSurf(0).GetMemorySize() + RhSurf(1).GetMemorySize() + RhSurf(2).GetMemorySize() 
      + RhSurf(3).GetMemorySize();
    
    taille += Seldon::GetMemorySize(num_dof_points_surf) + sizeof(*this);
    return taille;
  }
  
  //! constructs the finite element
  void ElementReference<Dimension2, 4>
  ::ConstructFiniteElement(int r, int rgeom, int rquad, int type_quad,
			   int rsurf, int type_surf)
  {
    this->fe_h1.ConstructFiniteElement(r);
    this->fe_hcurl.ConstructFiniteElement(r);

    this->elt_geom.quadrature_equal_nodal = false;
    this->elt_geom.dof_equal_nodal = false;
    this->elt_geom.dof_equal_quadrature = false;
    
    ConstructFunctions();
    ConstructMassMatrix();
  }
  
  
  //! construction of basis functions
  void ElementReference<Dimension2, 4>::ConstructFunctions()
  {
    int r = fe_h1.GetOrder();

    LocalScalNumber.Reallocate(fe_h1.GetNbDofBoundary(0), fe_h1.GetNbBoundaries());
    LocalVecNumber.Reallocate(fe_hcurl.GetNbDofBoundary(0), fe_h1.GetNbBoundaries());

    for (int n = 0; n < fe_h1.GetNbBoundaries(); n++)
      {
	for (int i = 0; i < fe_h1.GetNbDofBoundary(n); i++)
	  LocalScalNumber(i, n) = fe_h1.GetLocalNumber(n, i);

	for (int i = 0; i < fe_hcurl.GetNbDofBoundary(n); i++)
	  LocalVecNumber(i, n) = fe_hcurl.GetLocalNumber(n, i);
      }
    
    this->nb_dof_loc = fe_h1.GetNbDof() + fe_hcurl.GetNbDof();
    this->nb_dof_boundaries = 2*fe_h1.GetNbBoundaries()*r;
    
    // the dofs on the edges are constructed (with array EdgesDof)
    this->EdgesDof.Reallocate(2*r+1, fe_h1.GetNbBoundaries());    
    if (fe_h1.GetNbBoundaries() == 4)
      {
        // first, we number the four vertices
        this->EdgesDof(r, 0) = 0; this->EdgesDof(2*r, 0) = 1;
        this->EdgesDof(r, 1) = 1; this->EdgesDof(2*r, 1) = 2;
        this->EdgesDof(r, 2) = 2; this->EdgesDof(2*r, 2) = 3;
        this->EdgesDof(r, 3) = 3; this->EdgesDof(2*r, 3) = 0;
        
        // first edge
        int numero = 4;
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 0) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 0) = numero++;
        
        // second edge
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 1) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 1) = numero++;
        
        // third edge
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 2) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 2) = numero++;
        
        // fourth edge
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 3) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 3) = numero++;
        
        // dof numbers
        VectorialDof.Reallocate(fe_hcurl.GetNbDof());
        for (int num_loc = 0; num_loc < fe_h1.GetNbBoundaries(); num_loc++)
          for (int i = 0; i < r; i++)
            VectorialDof(num_loc*r + i) = this->EdgesDof(i, num_loc);
        
        for (int i = 4*r; i < fe_hcurl.GetNbDof(); i++)
          VectorialDof(i) = numero++;
        
        ScalarDof.Reallocate(fe_h1.GetNbDof());
        ScalarDof(0) = 0; ScalarDof(1) = 1;
        ScalarDof(2) = 2; ScalarDof(3) = 3;
        for (int num_loc = 0; num_loc < fe_h1.GetNbBoundaries(); num_loc++)
          for (int i = 0; i < r-1; i++)
            ScalarDof(4 + num_loc*(r-1) + i) = this->EdgesDof(i+r+1, num_loc);
        
        for (int i = 4*r; i < fe_h1.GetNbDof(); i++)
          ScalarDof(i) = numero++;
      }
    else
      {
        // first, we number the three vertices
        this->EdgesDof(r, 0) = 0; this->EdgesDof(2*r, 0) = 1;
        this->EdgesDof(r, 1) = 1; this->EdgesDof(2*r, 1) = 2;
        this->EdgesDof(r, 2) = 2; this->EdgesDof(2*r, 2) = 0;
        
        // first edge
        int numero = 3;
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 0) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 0) = numero++;
        
        // second edge
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 1) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 1) = numero++;
        
        // third edge
        for (int i = 0; i < r; i++)
          this->EdgesDof(i, 2) = numero++;
	
        for (int i = 1; i < r; i++)
          this->EdgesDof(i+r, 2) = numero++;
        
        // dof numbers
        VectorialDof.Reallocate(fe_hcurl.GetNbDof());
        for (int num_loc = 0; num_loc < fe_h1.GetNbBoundaries(); num_loc++)
          for (int i = 0; i < r; i++)
            VectorialDof(num_loc*r + i) = this->EdgesDof(i, num_loc);
        
        for (int i = 3*r; i < fe_hcurl.GetNbDof(); i++)
          VectorialDof(i) = numero++;
        
        ScalarDof.Reallocate(fe_h1.GetNbDof());
        ScalarDof(0) = 0; ScalarDof(1) = 1;
        ScalarDof(2) = 2;
        for (int num_loc = 0; num_loc < fe_h1.GetNbBoundaries(); num_loc++)
          for (int i = 0; i < r-1; i++)
            ScalarDof(3 + num_loc*(r-1) + i) = this->EdgesDof(i+r+1, num_loc);
        
        for (int i = 3*r; i < fe_h1.GetNbDof(); i++)
          ScalarDof(i) = numero++;
      }

    FillPositionDofBoundaries(EdgesDof, this->power_two_face, this->PosDofOnFace);        
    
    // pts-dof numbers (used to project a function on the finite element space)
    if ((fe_h1.GetNbPointsDofInside() != fe_h1.GetNbPointsDof())
	|| (fe_hcurl.GetNbPointsDofInside() != fe_hcurl.GetNbPointsDof()))
      {
	cout << "case not implemented" << endl;
	abort();
      }
    
    VectReal_wp points_dof1d;
    points_dof1d.Reallocate(2*r+1);
    for (int i = 0; i < r; i++)
      points_dof1d(i) = fe_hcurl.PointsDof1D(i);
    
    for (int i = 0; i <= r; i++)
      points_dof1d(r+i) = fe_h1.PointsDof1D(i);
    
    VectR2 points_dof2d;
    MergePoints(fe_hcurl.PointsDofND(), fe_h1.PointsDofND(),
		points_dof2d, VectorialPtsDof, ScalarPtsDof);

    this->nb_points_dof_inside = points_dof2d.GetM();
    
    this->num_dof_points_surf.Reallocate(fe_h1.GetNbBoundaries());
    Vector<int> IndexPointDof(points_dof2d.GetM());
    ScalarPtsDofFace.Reallocate(fe_h1.GetNbBoundaries());
    VectorialPtsDofFace.Reallocate(fe_h1.GetNbBoundaries());
    for (int n = 0; n < fe_h1.GetNbBoundaries(); n++)
      {
	// counting the number of points in the surface for dof projection
	IndexPointDof.Fill(-1);
	int nb_points = 0;
	ScalarPtsDofFace(n).Reallocate(fe_h1.GetNbPointsDofSurface(n));
	for (int i = 0; i < fe_h1.GetNbPointsDofSurface(n); i++)
	  {
	    int num = ScalarPtsDof(fe_h1.GetPointDofNumber(n, i));
	    if (IndexPointDof(num) == -1)
	      {
		IndexPointDof(num) = nb_points;
		nb_points++;
	      }
	    
	    ScalarPtsDofFace(n)(i) = IndexPointDof(num);
	  }
	
	VectorialPtsDofFace(n).Reallocate(fe_hcurl.GetNbPointsDofSurface(n));
	for (int i = 0; i < fe_hcurl.GetNbPointsDofSurface(n); i++)
	  {
	    int num = VectorialPtsDof(fe_hcurl.GetPointDofNumber(n, i));
	    if (IndexPointDof(num) == -1)
	      {
		IndexPointDof(num) = nb_points;
		nb_points++;
	      }
	    
	    VectorialPtsDofFace(n)(i) = IndexPointDof(num);
	  }
	
	this->num_dof_points_surf(n).Reallocate(nb_points);
	IndexPointDof.Fill(-1);
	nb_points = 0;
	for (int i = 0; i < fe_h1.GetNbPointsDofSurface(n); i++)
	  {
	    int num = ScalarPtsDof(fe_h1.GetPointDofNumber(n, i));
	    if (IndexPointDof(num) == -1)
	      {
		IndexPointDof(num) = nb_points;
		this->num_dof_points_surf(n)(nb_points) = num;
		nb_points++;
	      }
	  }

	for (int i = 0; i < fe_hcurl.GetNbPointsDofSurface(n); i++)
	  {
	    int num = VectorialPtsDof(fe_hcurl.GetPointDofNumber(n, i));
	    if (IndexPointDof(num) == -1)
	      {
		IndexPointDof(num) = nb_points;
		this->num_dof_points_surf(n)(nb_points) = num;
		nb_points++;
	      }
	  }
      }

    fe_h1.SetDofNumbersBoundary(num_dof_points_surf);
    
    this->SetPointsDof1D(points_dof1d);
    this->SetPointsDofND(points_dof2d);
    
    // quadrature points
    VectR2 points2d;
    //MergePoints(fe_hcurl.PointsQuadInsideND(), fe_h1.PointsQuadInsideND(),
    //		points2d, VectorialQuadrature, ScalarQuadrature);
    
    points2d = fe_h1.PointsQuadInsideND();
    VectorialQuadrature.Reallocate(points2d.GetM());
    ScalarQuadrature.Reallocate(points2d.GetM());
    VectorialQuadrature.Fill(); ScalarQuadrature.Fill();

    if (fe_hcurl.GetNbQuadBoundary(0) != fe_h1.GetNbQuadBoundary(0))
      {
	cout << "Quadrature points on the edges must coincide" << endl;
	abort();
      }
    
    for (int k = 0; k < fe_hcurl.GetNbQuadBoundary(0); k++)
      if (abs(fe_hcurl.PointsQuadratureBoundary(k, 0) - fe_h1.PointsQuadratureBoundary(k, 0)) > R2::threshold)
	{
	  cout << "Quadrature points on the edges must coincide" << endl;
	  abort();
	}
    
    nb_points_quadrature_inside = points2d.GetM();
    points2d_inside = points2d;
    
    points2d.Resize(points2d.GetM() + fe_h1.GetNbBoundaries()*fe_h1.GetNbQuadBoundary(0));
    int num = nb_points_quadrature_inside; R2 pt;
    num_quad_points_surf.Reallocate(fe_h1.GetNbBoundaries());
    for (int num_loc = 0; num_loc < fe_h1.GetNbBoundaries(); num_loc++)
      {
	num_quad_points_surf(num_loc).Reallocate(fe_h1.GetNbQuadBoundary(num_loc));
	for (int k = 0; k < fe_h1.GetNbQuadBoundary(num_loc); k++)
	  {
	    fe_h1.GetLocalCoordOnBoundary(num_loc, fe_h1.PointsQuadratureBoundary(k, num_loc), pt);
	    num_quad_points_surf(num_loc)(k) = num;
	    points2d(num++) = pt;
	  }
      }   
    
    this->SetPointsND(points2d);
    fe_h1.SetQuadNumbersBoundary(num_quad_points_surf);

    // and weights
    VectReal_wp weights2d(points2d.GetM());
    weights2d.Fill(0);
    for (int k = 0; k < ScalarQuadrature.GetM(); k++)
      weights2d(ScalarQuadrature(k)) = fe_h1.WeightsND(k);

    for (int k = 0; k < VectorialQuadrature.GetM(); k++)
      weights2d(VectorialQuadrature(k)) = fe_hcurl.WeightsND(k);

    this->SetWeightsND(weights2d);
    
    this->elt_geom.ComputeCoefficientTransformation();
    
    // finding dofs on the boundary
    /*pt_dof_on_boundary.Reallocate(points_dof2d.GetM(), fe_h1.GetNbBoundaries());
    pt_dof_on_boundary.Fill(true);
    
    if (fe_h1.GetNbBoundaries() == 4)
      {
	pt_dof_on_boundary.Fill(false);
	for (int i = 0; i < points_dof2d.GetM(); i++)
	  {
	    if (abs(points_dof2d(i)(1)) <= R2::threshold)
	      pt_dof_on_boundary(i, 0) = true;

	    if (abs(1.0-points_dof2d(i)(0)) <= R2::threshold)
	      pt_dof_on_boundary(i, 1) = true;

	    if (abs(1.0-points_dof2d(i)(1)) <= R2::threshold)
	      pt_dof_on_boundary(i, 2) = true;

	    if (abs(points_dof2d(i)(0)) <= R2::threshold)
	      pt_dof_on_boundary(i, 3) = true;
	  }
      }
    */    
    //DISP(ScalarDof); DISP(VectorialDof);
    //DISP(ScalarPtsDof); DISP(VectorialPtsDof);
    //DISP(this->EdgesDof);
    //DISP(this->points_dof2d);
  }
  
  
  //! construction of mass matrix
  void ElementReference<Dimension2, 4>::ConstructMassMatrix()
  {
    ChE.Reallocate(fe_hcurl.GetNbDof(), 2*this->nb_points_quadrature_inside);
    ChH.Reallocate(this->GetNbDofH(), 2*this->nb_points_quadrature_inside);
    ChEteta.Reallocate(fe_h1.GetNbDof(), this->nb_points_quadrature_inside);
    ChHteta.Reallocate(fe_hcurl.GetNbDofH(), this->nb_points_quadrature_inside);
    
    ChE.Fill(0); ChH.Fill(0);
    ChEteta.Fill(0); ChHteta.Fill(0);

    RhE.Reallocate(fe_hcurl.GetNbDof(), this->nb_points_quadrature_inside);
    RhEteta.Reallocate(fe_h1.GetNbDof(), 2*this->nb_points_quadrature_inside);
    RhE.Fill(0); RhEteta.Fill(0);
    
    VectR2 phi, grad_psi; VectReal_wp psi, curl_phi;
    for (int i = 0; i < this->nb_points_quadrature_inside; i++)
      {
        fe_hcurl.ComputeValuesPhiRef(this->PointsND(i), phi);
        for (int j = 0; j < fe_hcurl.GetNbDof(); j++)
          {
	    ChE(j, 2*i) = phi(j)(0);
	    ChE(j, 2*i+1) = phi(j)(1);
          }
        
        fe_h1.ComputeValuesPhiRef(this->PointsND(i), psi);
        for (int j = 0; j < fe_h1.GetNbDof(); j++)
          {
	    ChH(2*j, 2*i) = psi(j);
	    ChH(2*j+1, 2*i+1) = psi(j);
	    ChEteta(j, i) = psi(j);
          }
        
        fe_hcurl.ComputeValuesPhiHRef(this->PointsND(i), psi);
        for (int j = 0; j < fe_hcurl.GetNbDofH(); j++)
	  ChHteta(j, i) = psi(j);
        
        fe_hcurl.ComputeCurlPhiRef(this->PointsND(i), curl_phi);
        for (int j = 0; j < fe_hcurl.GetNbDof(); j++)
	  RhE(j, i) = curl_phi(j);

        fe_h1.ComputeGradientPhiRef(this->PointsND(i), grad_psi);
        for (int j = 0; j < fe_h1.GetNbDof(); j++)
	  {
	    RhEteta(j, 2*i) = grad_psi(j)(0);
	    RhEteta(j, 2*i+1) = grad_psi(j)(1);
	  }
      }
    
    int r = fe_h1.GetOrder();
    Globatto<Real_wp> gauss;
    gauss.ConstructQuadrature(r, gauss.QUADRATURE_GAUSS);
    gauss.ComputeGradPhi(1e3*epsilon_machine);
    RhSurf(0).Reallocate(2*(r+1), r+1);
    RhSurf(1).Reallocate(2*(r+1), r+1);
    RhSurf(2).Reallocate(2*(r+1), r+1);
    RhSurf(3).Reallocate(2*(r+1), r+1);
    RhSurf(0).Fill(0); RhSurf(1).Fill(0);
    RhSurf(2).Fill(0); RhSurf(3).Fill(0);
    for (int i = 0; i <= r; i++)
      for (int j = 0; j <= r; j++)
        {
          Real_wp dphi = gauss.GradPhi(i, j);
          RhSurf(0)(2*j, i) = dphi;          
          RhSurf(1)(2*j+1, i) = dphi;          
          RhSurf(2)(2*j, i) = -dphi;
          RhSurf(3)(2*j+1, i) = -dphi;
        }
  }
  
  
  void ElementReference<Dimension2, 4>
  ::MergePoints(const VectR2& ptsA, const VectR2& ptsB, VectR2& pts, IVect& numA, IVect& numB)
  {
    int nbA = ptsA.GetM(), nbB = ptsB.GetM();
    VectR2 points_sort(ptsA);
    points_sort.PushBack(ptsB);

    IVect permut(nbA+nbB);
    permut.Fill();
    
    Sort(points_sort, permut);

    int nb = 1;
    R2 point_prev(points_sort(0));
    pts.Reallocate(nbA+nbB);
    pts(0) = points_sort(0);
    numA.Reallocate(nbA);
    numB.Reallocate(nbB);
    numA.Fill(-1); numB.Fill(-1);
    for (int k = 0; k < points_sort.GetM(); k++)
      {
	if (points_sort(k) != point_prev)
	  {
	    pts(nb) = points_sort(k);
	    nb++;
	  }
	
	if (permut(k) < nbA)
	  numA(permut(k)) = nb-1;
	else
	  numB(permut(k)-nbA) = nb-1;

	point_prev = points_sort(k);
      }
    
    pts.Resize(nb);
  }
  
  
  //! returns the values of basis functions associated with Hrz on a point
  /*!
    \param[in] point point where basis functions are evaluated
    \param[out] phi values of basis functions associated with Hrz
   */
  void ElementReference<Dimension2, 4>
  ::ComputeValuesPhiHRef(const R2& point, VectR2& phi) const
  {
    phi.Reallocate(GetNbDofH());
    VectReal_wp psi;
    fe_h1.ComputeValuesPhiRef(point, psi);
    for (int i = 0; i < psi.GetM(); i++)
      {
        phi(2*i)(0) = psi(i);
        phi(2*i)(1) = 0.0;
        phi(2*i+1)(0) = 0.0;
        phi(2*i+1)(1) = psi(i);
      }
  }
  
  
  //! computes values of basis functions on a point
  /*!
    \param[in] point_loc point where basis functions are evaluated 
    \param[in] point_glob global points where basis functions are evaluated
    \param[out] res values of basis functions in 3-D : (r, theta, z) components
    \param[in] dfjm1 inverse of DF_j
    \param[in] FaceBasis finite element considered
    \param[in] mesh geometric mesh
    \param[in] mesh_num object containing mesh dofs
    \param[in] nquad element number
    Piola transform DF_j^*-1 is applied to Erz dofs
   */
  void ElementReference<Dimension2, 4>::
  ComputeValuesPhi(const R2& point_loc, const R2& point_glob, Vector<R3>& res,
		   const Matrix2_2& dfjm1,
		   const ElementReference<Dimension2, 4>& FaceBasis,
                   const Mesh<Dimension2>& mesh,
		   const MeshNumbering<Dimension2>& mesh_num, int nquad) const
  {
    VectReal_wp phi_scal; VectR2 phi_vec;
    fe_hcurl.ComputeValuesPhiRef(point_loc, phi_vec);
    fe_h1.ComputeValuesPhiRef(point_loc, phi_scal);
    res.Reallocate(this->GetNbDof());
    // scalar dofs
    for (int i = 0; i < phi_scal.GetM(); i++)
      {
	int num_dof_loc = ScalarDof(i);
	res(num_dof_loc)(0) = 0;
	res(num_dof_loc)(1) = phi_scal(i);
	res(num_dof_loc)(2) = 0;
      }
    
    // vectorial dofs (transformation DF_i^{*-1})
    R2 phi, vec;
    for (int i = 0; i < phi_vec.GetM(); i++)
      {
	int num_dof_loc = VectorialDof(i);
	phi = phi_vec(i);
	MltTrans(dfjm1, phi, vec);
	
	res(num_dof_loc)(0) = vec(0);
        res(num_dof_loc)(1) = 0.0;
        res(num_dof_loc)(2) = vec(1);
      }
    
    // signs to take into account
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, nquad);
  }
  
  
  //! computes curl of basis functions on a point
  /*!
    \param[in] point_loc point where basis functions are evaluated 
    \param[in] point_glob global points where basis functions are evaluated
    \param[out] res curl of basis functions in 3-D : (r, theta, z) components
    \param[in] dfjm1 inverse of DF_j
    \param[in] FaceBasis finite element considered
    \param[in] mesh geometric mesh
    \param[in] mesh_num object containing mesh dofs
    \param[in] nquad element number
  */
  void ElementReference<Dimension2, 4>::
  ComputeValuesGradientPhi(const R2& point_loc, const R2& point_glob,
			   Vector<R3>& res, const Matrix2_2& dfjm1,
			   const ElementReference<Dimension2, 4>& FaceBasis,
			   const Mesh<Dimension2>& mesh,
			   const MeshNumbering<Dimension2>& mesh_num, int nquad) const
  {
    VectR2 grad_scal; VectReal_wp curl_vec; R2 grad;
    fe_h1.ComputeGradientPhiRef(point_loc, grad_scal);
    fe_hcurl.ComputeCurlPhiRef(point_loc, curl_vec);
    res.Reallocate(GetNbDof());
    // scalar dofs
    for (int i = 0; i < grad_scal.GetM(); i++)
      {
	int num_dof_loc = ScalarDof(i);
	MltTrans(dfjm1, grad_scal(i), grad);
	res(num_dof_loc)(0) = grad(1);
	res(num_dof_loc)(1) = 0;
	res(num_dof_loc)(2) = -grad(0);
      }
    
    // vectorial dofs (transformation 1/Ji)
    Real_wp invJi = Det(dfjm1);
    for (int i = 0; i < curl_vec.GetM(); i++)
      {
	int num_dof_loc = VectorialDof(i);
	res(num_dof_loc)(0) = 0.0;
        res(num_dof_loc)(1) = curl_vec(i)*invJi;
        res(num_dof_loc)(2) = 0.0;	
      }
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, nquad);
  }
  
  
  //! not implemented
  template<class T>
  void ElementReference<Dimension2, 4>::
  GetCurlFromGradient(const Vector<Vector<T> >& dU,
                      Vector<Vector<T> >& curlU)
  {
    cout << "Not implemented for this finite element" << endl;
    abort();
  }
  
  
  //! computes values of basis functions on quadrature points of an edge
  /*!
    \param[out] ValuePhi_Boundary values of basis functions on quadrature points of the edge
    \param[in] PointsElem object containing position of points after transform F_j
    \param[in] MatricesElem object containing jacobian matrices DF_j
    \param[in] FaceBasis finite element considered
    \param[in] mesh geometric mesh
    \param[in] mesh_num object containing mesh dofs
    \param[in] nquad element number
    \param[in] num_loc edge number
    Piola transform DF_j^*-1 is applied to Erz dofs
   */
  void ElementReference<Dimension2, 4>::
  ComputeValuesPhiBoundary(Matrix<R3>& ValuePhi_Boundary,
			   const SetPoints<Dimension2>& PointsElem,
			   const SetMatrices<Dimension2>& MatricesElem,
                           const ElementReference<Dimension2, 4>& FaceBasis,
			   const Mesh<Dimension2>& mesh,
                           const MeshNumbering<Dimension2>& mesh_num,
			   int nquad, int num_loc) const
  {
    int nb_quad = this->GetNbQuadBoundary(num_loc);
    ValuePhi_Boundary.Reallocate(GetNbDof(), nb_quad);
    FillZero(ValuePhi_Boundary);

    Matrix2_2 dfjm1; R2 val_phi;
    VectReal_wp phi; VectR2 phi_vec;
    for (int k = 0; k < nb_quad; k++)
      {
        int num_point = this->GetQuadNumber(num_loc, k);
        fe_h1.GetValuePhiOnQuadraturePoint(num_point, phi);
        for (int i = 0; i < phi.GetM(); i++)
          {
            int num_dof = ScalarDof(i);
            ValuePhi_Boundary(num_dof, k)(1) = phi(i);
          }
        
        GetInverse(MatricesElem.GetPointQuadratureBoundary(k), dfjm1);
        fe_hcurl.GetValuePhiOnQuadraturePoint(num_point, phi_vec);
        for (int i = 0; i < FaceBasis.fe_hcurl.GetNbDof(); i++)
          {
            int num_dof = VectorialDof(i);
            MltTrans(dfjm1, phi_vec(i), val_phi);            
	    ValuePhi_Boundary(num_dof, k)(0) = val_phi(0); 
            ValuePhi_Boundary(num_dof, k)(2) = val_phi(1); 
          }
      }
    
    // we take into account signs of basis functions 
    mesh_num.number_map.ModifyLocalRowMatrix(mesh_num, ValuePhi_Boundary, nquad);
  }
  
  
  //! computes curl of basis functions on quadrature points of an edge
  /*!
    \param[out] GradientPhi_Boundary curls of basis functions on quadrature points of the edge
    \param[in] PointsElem object containing position of points after transform F_j
    \param[in] MatricesElem object containing jacobian matrices DF_j
    \param[in] FaceBasis finite element considered
    \param[in] mesh geometric mesh
    \param[in] mesh_num object containing mesh dofs
    \param[in] nquad element number
    \param[in] num_loc edge number
   */
  void ElementReference<Dimension2, 4>::
  ComputeValuesGradientPhiBoundary(Matrix<R3>& GradientPhi_Boundary,
                                   const SetPoints<Dimension2>& PointsElem,
                                   const SetMatrices<Dimension2>& MatricesElem,
                                   const ElementReference<Dimension2, 4>& FaceBasis,
                                   const Mesh<Dimension2>& mesh,
				   const MeshNumbering<Dimension2>& mesh_num, int nquad, int num_loc) const
  {
    int nb_quad = this->GetNbQuadBoundary(num_loc);
    GradientPhi_Boundary.Reallocate(this->GetNbDof(), nb_quad);
    FillZero(GradientPhi_Boundary);

    Matrix2_2 dfjm1; VectReal_wp curl_phi;
    VectR2 grad_psi; R2 grad; Real_wp invJacob;
    for (int k = 0; k < nb_quad; k++)
      {
        int num_point = this->GetQuadNumber(num_loc, k);
        GetInverse(MatricesElem.GetPointQuadratureBoundary(k), dfjm1);
        fe_h1.GetGradientPhiOnQuadraturePoint(num_point, grad_psi);
        for (int i = 0; i < grad_psi.GetM(); i++)
          {
            int num_dof = ScalarDof(i);
	    MltTrans(dfjm1, grad_psi(i), grad);
            GradientPhi_Boundary(num_dof, k)(0) = grad(1);
            GradientPhi_Boundary(num_dof, k)(2) = -grad(0);
          }
        
        fe_hcurl.GetCurlPhiOnQuadraturePoint(num_point, curl_phi);
	invJacob = 1.0/Det(MatricesElem.GetPointQuadratureBoundary(k));
        for (int i = 0; i < FaceBasis.fe_hcurl.GetNbDof(); i++)
          {
            int num_dof = VectorialDof(i);
	    GradientPhi_Boundary(num_dof, k)(1) = curl_phi(i)*invJacob; 
          }
      }
    
    // we take into account signs of basis functions 
    mesh_num.number_map.ModifyLocalRowMatrix(mesh_num, GradientPhi_Boundary, nquad);
  }
  
  
  //! not implemented
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeValueNodalBoundary(const Vector<T>& Uloc, Vector<T>& Uboundary,
			    const ElementReference<Dimension2, 4>& FaceBasis,
                            const Mesh<Dimension2>& mesh, int n, int num_loc) const
  {
    cout << "Not implemented for this finite element" << endl;
    abort();
  }
  

  //! not implemented
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeQuadratureValues(const SetPoints<Dimension2>& PointsElem,
			  const SetMatrices<Dimension2>& MatricesElem,
			  const Vector<Vector<T> >& Uloc, Vector<Vector<T> >& Uloc_node,
			  Vector<Vector<T> >& dUloc_node, bool compute_u, bool compute_curl,
			  const ElementReference<Dimension2, 4>& FaceBasis,
			  const Mesh<Dimension2>& mesh, int iquad) const
  {
    cout << "Not implemented for this finite element" << endl;
    abort();
  }
  
  
  //! to be done
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeNodalValues(const SetPoints<Dimension2>& PointsElem,
		     const SetMatrices<Dimension2>& MatricesElem,
		     const Vector<Vector<T> >& Uloc, Vector<Vector<T> >& Uloc_node,
                     const ElementReference<Dimension2, 4>& FaceBasis,
		     const Mesh<Dimension2>& mesh, int i) const
  {
    cout << "Not implemented for this finite element" << endl;
    abort();
  }
  

  //! not implemented  
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeValueBoundary(const SetPoints<Dimension2>& pts,
                       const SetMatrices<Dimension2>& mat,
                       const Vector<Vector<T> >& feval, Vector<Vector<TinyVector<T, 3> > >& res,
                       const ElementReference<Dimension2, 4>& FaceBasis,
                       const Mesh<Dimension2>& mesh, int n, int num_loc) const
  {
    cout << "Not implemented for this finite element" << endl;
    abort();
  }
  

  //! not implemented  
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeGradientBoundary(const SetPoints<Dimension2>& pts,
                          const SetMatrices<Dimension2>& mat,
                          const Vector<Vector<T> >& feval, Vector<Vector<TinyVector<T, 3> > >& res,
                          const ElementReference<Dimension2, 4>& FaceBasis,
                          const Mesh<Dimension2>& mesh, int n, int num_loc) const
  {
    cout << "Not implemented for this finite element" << endl;
    abort();
  }
  
  
  //! project function f on degrees of freedom
  /*!
    \param[in] PointsElem reference points on the quad
    \param[in] MatricesElem jacobian matrices
    \param[in] feval values of the function f on dof points
    \param[out] res components on degrees of freedom
    \param[in] vars given problem
    \param[in] FaceBasis given finite element
    \param[in] n element number
  */
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeProjectionDof(const SetPoints<Dimension2>& PointsElem,
		       const SetMatrices<Dimension2>& MatricesElem,
		       Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
		       const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
		       const MeshNumbering<Dimension2>& mesh_num, int n) const
  {
    int nb_dof_loc = this->GetNbDof();
    if (feval.GetM() != 1)
      {
	abort();
      }
    
    res.Reallocate(1);
    res(0).Reallocate(nb_dof_loc);
    
    TinyVector<T, 2> phi, fvec;
    Vector<T> feval_h1(ScalarPtsDof.GetM()), proj_h1(fe_h1.GetNbDof());
    Vector<T> feval_hcurl(2*fe_hcurl.GetNbPointsDof()), proj_hcurl(fe_hcurl.GetNbDof());
    
    // we get values on nodal points, multiplication by DF_i^* for vectorial dofs
    for (int i = 0; i < ScalarPtsDof.GetM(); i++)
      {
	int node = ScalarPtsDof(i);
	feval_h1(i) = feval(0)(3*node+1);
      }
    
    fe_h1.ComputeProjectionDofRef(feval_h1, proj_h1);
    
    for (int i = 0; i < fe_hcurl.GetNbPointsDof(); i++)
      {
	int node = VectorialPtsDof(i);
	phi(0) = feval(0)(3*node);
	phi(1) = feval(0)(3*node+2);
	MltTrans(MatricesElem.GetPointDof(node), phi, fvec);
	
	feval_hcurl(2*i) = fvec(0);
	feval_hcurl(2*i+1) = fvec(1);
      }
    
    fe_hcurl.ComputeProjectionDofRef(feval_hcurl, proj_hcurl);
    
    for (int i = 0; i < proj_h1.GetM(); i++)
      res(0)(ScalarDof(i)) = proj_h1(i);
    
    for (int i = 0; i < proj_hcurl.GetM(); i++)
      res(0)(VectorialDof(i)) = proj_hcurl(i);
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, n);
  }


  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeProjectionSurfaceDof(const SetPoints<Dimension2>& PointsElem,
			      const SetMatrices<Dimension2>& MatricesElem,
			      Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
			      const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
			      const MeshNumbering<Dimension2>& mesh_num, int n, int num_loc) const
  {
    int nb_dof_loc = this->GetNbDofBoundary(num_loc);
    if (feval.GetM() != 1)
      {
	abort();
      }
    
    res.Reallocate(1);
    res(0).Reallocate(nb_dof_loc);
    
    TinyVector<T, 2> phi, fvec;
    Vector<T> feval_h1(ScalarPtsDofFace(num_loc).GetM()), proj_h1(fe_hcurl.GetOrder()+1);
    Vector<T> feval_hcurl(2*fe_hcurl.GetNbPointsDofSurface(num_loc)), proj_hcurl(fe_hcurl.GetNbDof());
    
    // we get values on nodal points, multiplication by DF_i^* for vectorial dofs
    for (int i = 0; i < ScalarPtsDofFace(num_loc).GetM(); i++)
      {
	int node = ScalarPtsDofFace(num_loc)(i);
	feval_h1(i) = feval(0)(3*node+1);
      }
    
    fe_h1.ComputeProjectionSurfaceDofRef(feval_h1, proj_h1, num_loc);
    
    for (int i = 0; i < fe_hcurl.GetNbPointsDofSurface(num_loc); i++)
      {
	int node = VectorialPtsDofFace(num_loc)(i);
	phi(0) = feval(0)(3*node);
	phi(1) = feval(0)(3*node+2);
	MltTrans(MatricesElem.GetPointDofBoundary(node), phi, fvec);
	
	feval_hcurl(2*i) = fvec(0);
	feval_hcurl(2*i+1) = fvec(1);
      }
    
    fe_hcurl.ComputeProjectionSurfaceDofRef(feval_hcurl, proj_hcurl, num_loc);

    Vector<Vector<T> > u_loc(1);
    u_loc(0).Reallocate(FaceBasis.GetNbDof());
    u_loc(0).Zero();
    
    for (int i = 0; i < proj_h1.GetM(); i++)
      u_loc(0)(ScalarDof(LocalScalNumber(i, num_loc))) = proj_h1(i);
    
    for (int i = 0; i < proj_hcurl.GetM(); i++)
      u_loc(0)(VectorialDof(LocalVecNumber(i, num_loc))) = proj_hcurl(i);
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, u_loc, n);
    
    for (int i = 0; i < nb_dof_loc; i++)
      res(0)(i) = u_loc(0)(this->EdgesDof(i, num_loc));
  }
  
  
  //! computation of \f$ \int_{\hat{K}} f \varphi_i \f$
  /*!
    \param[in] pts reference points on the quad
      \param[in] mat jacobian matrices
      \param[in] feval values of the function f on quadrature points
      \param[out] res integral agains each basis function
      \param[in] vars given problem
      \param[in] FaceBasis given finite element
      \param[in] nquad element number
  */
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeIntegral(const SetPoints<Dimension2>& pts,
		  const SetMatrices<Dimension2>& mat,
		  Vector<Vector<T> > & feval, Vector<Vector<T> > & res,
		  const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
		  const MeshNumbering<Dimension2>& mesh_num, int nquad) const
  {
    int nb_dof_loc = this->GetNbDof();
    int nb_points_h1 = fe_h1.GetNbPointsQuadratureInside();
    int nb_points_hcurl = fe_hcurl.GetNbPointsQuadratureInside();
    
    Vector<T> feval_h1(nb_points_h1), res_h1(fe_h1.GetNbDof());
    Vector<T> feval_hcurl(2*nb_points_hcurl), res_hcurl(fe_hcurl.GetNbDof());

    if (feval.GetM() != 1)
      {
	abort();
      }
    
    res.Reallocate(1);    
    res(0).Reallocate(nb_dof_loc);
    
    Matrix2_2 dfjm1;
    TinyVector<T, 2> phi, fvec;
    
    // application of transformation DF_i^{-1} for vectorial dofs
    for (int i = 0; i < nb_points_hcurl; i++)
      {
	int node = VectorialQuadrature(i);
	phi(0) = feval(0)(3*node);
	phi(1) = feval(0)(3*node+2);
	GetInverse(mat.GetPointQuadrature(node), dfjm1);
	Mlt(dfjm1, phi, fvec);
	Real_wp poids = fe_hcurl.WeightsND(i)*Det(mat.GetPointQuadrature(node));
	feval_hcurl(2*i) = fvec(0)*poids;
	feval_hcurl(2*i+1) = fvec(1)*poids;
      }
    
    for (int i = 0; i < nb_points_h1; i++)
      {
	int node = ScalarQuadrature(i);
	Real_wp poids = fe_h1.WeightsND(node)*Det(mat.GetPointQuadrature(node));
	feval_h1(i) = poids*feval(0)(3*node+1);
      }
    
    // scalar dofs
    fe_h1.ComputeIntegralRef(feval_h1, res_h1);
    fe_hcurl.ComputeIntegralRef(feval_hcurl, res_hcurl);
    
    for (int i = 0; i < res_h1.GetM(); i++)
      res(0)(ScalarDof(i)) = res_h1(i);
    
    for (int i = 0; i < res_hcurl.GetM(); i++)
      res(0)(VectorialDof(i)) = res_hcurl(i);
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, nquad);
  }
  
  
  //! computation of \f$ \int_{\hat{K}} f \nabla \varphi \f$
  /*!
    \param[in] pts reference points on the quad
    \param[in] mat jacobian matrices
    \param[in] feval values of the function f on quadrature points
    \param[out] res integral agains gradient of each basis function
    \param[in] vars given problem
    \param[in] FaceBasis given finite element
    \param[in] nquad element number
    to be done
  */
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeIntegralGradient(const SetPoints<Dimension2>& pts,
			  const SetMatrices<Dimension2>& mat,
			  Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
			  const ElementReference<Dimension2, 4>& FaceBasis,
			  const Mesh<Dimension2>& mesh,
			  const MeshNumbering<Dimension2>& mesh_num, int nquad) const
  {
    int nb_dof_loc = GetNbDof();
    int nb_points_h1 = fe_h1.GetNbPointsQuadratureInside();
    int nb_points_hcurl = fe_hcurl.GetNbPointsQuadratureInside();

    if (feval.GetM() != 1)
      {
	abort();
      }
    
    res.Reallocate(1);    
    res(0).Reallocate(nb_dof_loc);
    
    Vector<T> feval_h1(2*nb_points_h1), res_h1(fe_h1.GetNbDof());
    Vector<T> feval_hcurl(nb_points_hcurl), res_hcurl(fe_hcurl.GetNbDof());
        
    Matrix2_2 dfjm1;
    TinyVector<T, 2> phi, fvec;
    
    // application of transformation DF_i^{-1} for scalar dofs
    for (int i = 0; i < nb_points_h1; i++)
      {
	int node = ScalarQuadrature(i);
	phi(0) = -feval(0)(3*node+2);
	phi(1) = feval(0)(3*node);
	GetInverse(mat.GetPointQuadrature(node), dfjm1);
	Mlt(dfjm1, phi, fvec);
	Real_wp poids = fe_h1.WeightsND(node)*Det(mat.GetPointQuadrature(node));
	feval_h1(2*i) = fvec(0)*poids;
	feval_h1(2*i+1) = fvec(1)*poids;
      }
    
    for (int i = 0; i < nb_points_hcurl; i++)
      {
	int node = VectorialQuadrature(i);
	Real_wp poids = fe_hcurl.WeightsND(i);
	feval_hcurl(i) = poids*feval(0)(3*node+1);
      }
    
    // scalar dofs
    fe_h1.ComputeIntegralGradientRef(feval_h1, res_h1);
    fe_hcurl.ComputeIntegralCurlRef(feval_hcurl, res_hcurl);
    
    for (int i = 0; i < res_h1.GetM(); i++)
      res(0)(ScalarDof(i)) = res_h1(i);
    
    for (int i = 0; i < res_hcurl.GetM(); i++)
      res(0)(VectorialDof(i)) = res_hcurl(i);
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, nquad);
  }
  
  
  //! computation of \f$ \int_{\hat{K}} f \varphi_i \f$
  /*!
    \param[in] pts reference points on the quad
    \param[in] mat jacobian matrices
    \param[in] feval values of the function f on quadrature points
    \param[out] contribH integral against basis functions associated with Hrz
    \param[out] contribHteta integral against basis functions associated with Htheta
    \param[in] vars given problem
    \param[in] FaceBasis given finite element
    \param[in] nquad element number
    Here varphi_i designs a basis function associated with Hrz or Htheta
  */
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeIntegralH(const SetPoints<Dimension2>& pts,
                   const SetMatrices<Dimension2>& mat,
                   Vector<T>& feval, Vector<T>& contribH,
                   Vector<T>& contribHteta, const ElementReference<Dimension2, 4>& FaceBasis,
                   const Mesh<Dimension2>& mesh, int nquad) const
  {
    int nb_points_h1 = fe_h1.GetNbPointsQuadratureInside();
    int nb_points_hcurl = fe_hcurl.GetNbPointsQuadratureInside();
    
    contribH.Reallocate(FaceBasis.GetNbDofH());
    contribHteta.Reallocate(FaceBasis.GetNbDofHteta());
    
    Matrix2_2 dfjm1;
    TinyVector<T, 2> phi, fvec;
    Vector<T> feval_H(2*nb_points_hcurl);
    Vector<T> feval_Hteta(nb_points_h1);
    for (int j = 0; j < nb_points_hcurl; j++)
      {
        int node = VectorialQuadrature(j);
	//int node = j;
        phi(0) = feval(3*node);
        phi(1) = feval(3*node+2);
        GetInverse(mat.GetPointQuadrature(node), dfjm1);
        Mlt(dfjm1, phi, fvec);
        Real_wp radius = pts.GetPointQuadrature(node)(0);
        Real_wp poids = radius*fe_hcurl.WeightsND(j)*Det(mat.GetPointQuadrature(node));
        feval_H(2*node) = fvec(0)*poids;
        feval_H(2*node+1) = fvec(1)*poids;
      }
    
    for (int i = 0; i < nb_points_h1; i++)
      {
        int node = ScalarQuadrature(i);
	//int node = i;
        Real_wp radius = pts.GetPointQuadrature(node)(0);
        Real_wp poids = radius*fe_h1.WeightsND(node)*Det(mat.GetPointQuadrature(node));
        feval_Hteta(node) = poids*feval(3*node+1);
      }
    
    ApplyChHteta(feval_Hteta, contribHteta);
    ApplyChH(feval_H, contribH);
  }
  

  //! computation of \f$ \int_\Sigma f \phi_i ds \f$
  /*!
    \param[in] pts reference points on the quad
    \param[in] mat jacobian matrices
    \param[in] feval values of the function f on quadrature points of the edge
    \param[out] res integral agains each basis function
    \param[in] vars given problem
    \param[in] FaceBasis given finite element
    \param[in] n element number
    \param[in] num_loc local edge number
  */
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeIntegralSurface(const SetPoints<Dimension2>& pts,
			 const SetMatrices<Dimension2>& mat,
			 Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
			 const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
			 const MeshNumbering<Dimension2>& mesh_num, int n, int num_loc) const
  {
    int nb_dof_loc = GetNbDof();
    int nb_points_quad = GetNbQuadBoundary(num_loc);

    if (feval.GetM() != 1)
      {
	abort();
      }
    
    res.Reallocate(1);    
    res(0).Reallocate(nb_dof_loc);
    
    Vector<T> feval_h1(nb_points_quad), res_h1(fe_h1.GetNbDof());
    Vector<T> feval_hcurl(2*nb_points_quad), res_hcurl(fe_hcurl.GetNbDof());
    
    Matrix2_2 dfjm1;
    TinyVector<T, 2> phi, fvec;
    // application of transformation DF_i^{-1} for vectorial dofs
    for (int i = 0; i < nb_points_quad; i++)
      {
	phi(0) = feval(0)(3*i);
	phi(1) = feval(0)(3*i+2);
	GetInverse(mat.GetPointQuadratureBoundary(i), dfjm1);
	Mlt(dfjm1, phi, fvec);
	Real_wp poids = this->Weights1D(i)*mat.GetDsQuadratureBoundary(i);
	feval_hcurl(2*i) = fvec(0)*poids;
	feval_hcurl(2*i+1) = fvec(1)*poids;
	feval_h1(i) = poids*feval(0)(3*i+1);
      }
    
    // scalar dofs
    fe_h1.ComputeIntegralSurfaceRef(feval_h1, res_h1, num_loc);
    fe_hcurl.ComputeIntegralSurfaceRef(feval_hcurl, res_hcurl, num_loc);
    
    for (int i = 0; i < res_h1.GetM(); i++)
      res(0)(ScalarDof(i)) = res_h1(i);

    for (int i = 0; i < res_hcurl.GetM(); i++)
      res(0)(VectorialDof(i)) = res_hcurl(i);
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, n);    
  }
  
  
  //! integration against curl of basis functions on an edge of the mesh
  template<class T>
  void ElementReference<Dimension2, 4>::
  ComputeIntegralSurfaceGradient(const SetPoints<Dimension2>& pts,
				 const SetMatrices<Dimension2>& mat,
				 Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
				 const ElementReference<Dimension2, 4>& FaceBasis,
				 const Mesh<Dimension2>& mesh,
				 const MeshNumbering<Dimension2>& mesh_num, int n, int num_loc) const
  {
    int nb_dof_loc = this->GetNbDof();
    int nb_points_quad = this->GetNbQuadBoundary(num_loc);

    if (feval.GetM() != 1)
      {
	abort();
      }
    
    res.Reallocate(1);    
    res(0).Reallocate(nb_dof_loc);
    
    Vector<T> feval_h1(2*nb_points_quad), res_h1(fe_h1.GetNbDof());
    Vector<T> feval_hcurl(nb_points_quad), res_hcurl(fe_hcurl.GetNbDof());
        
    Matrix2_2 dfjm1; Real_wp invJacob;
    TinyVector<T, 2> phi, fvec;
    
    // application of transformation DF_i^{-1} for scalar dofs
    for (int i = 0; i < nb_points_quad; i++)
      {
	phi(0) = -feval(0)(3*i+2);
	phi(1) = feval(0)(3*i);
	GetInverse(mat.GetPointQuadratureBoundary(i), dfjm1);
	Mlt(dfjm1, phi, fvec);
	invJacob = 1.0/Det(mat.GetPointQuadratureBoundary(i));
	Real_wp poids = FaceBasis.Weights1D(i)*mat.GetDsQuadratureBoundary(i);
	feval_h1(2*i) = fvec(0)*poids;
	feval_h1(2*i+1) = fvec(1)*poids;
	feval_hcurl(i) = poids*feval(0)(3*i+1)*invJacob;
      }
    
    // integration against curl of basis functions
    fe_h1.ComputeIntegralSurfaceGradientRef(feval_h1, res_h1, num_loc);
    fe_hcurl.ComputeIntegralSurfaceCurlRef(feval_hcurl, res_hcurl, num_loc);
    
    for (int i = 0; i < res_h1.GetM(); i++)
      res(0)(ScalarDof(i)) = res_h1(i);
    
    for (int i = 0; i < res_hcurl.GetM(); i++)
      res(0)(VectorialDof(i)) = res_hcurl(i);
    
    mesh_num.number_map.ModifyLocalUnknownVector(mesh_num, res, n);    
  }


  void ElementReference<Dimension2, 4>::ApplyRhQuadratureTranspose(const VectReal_wp&, VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
  
  
  void ElementReference<Dimension2, 4>::ApplyRhTranspose(const VectReal_wp&, VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }    
  
  
  void ElementReference<Dimension2, 4>::ApplyChTranspose(const VectReal_wp&, VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }    
  
  
  void ElementReference<Dimension2, 4>::ApplyCh(const VectReal_wp&, VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort();
  }    
  
  
  void ElementReference<Dimension2, 4>::MltMassMatrix(VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
  
  
  void ElementReference<Dimension2, 4>::SolveMassMatrix(VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
  
  
  void ElementReference<Dimension2, 4>::SolveCholesky(const SeldonTranspose&, VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
    
  
  void ElementReference<Dimension2, 4>
  ::ComputeMassMatrix(VirtualMatrix<Real_wp>&, const VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
  
  
  void ElementReference<Dimension2, 4>
  ::IntegrateMassMatrix(const VirtualMatrix<Real_wp>&, const VectReal_wp&) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
  
  template<class T>
  void ElementReference<Dimension2, 4>
  ::ComputeGaussIntegralSurface(const Vector<Matrix2_2>& dfjm1,
				const VectReal_wp& weights, const VectReal_wp& ds,
				Vector<Vector<T> > & feval, Vector<Vector<T> >& res,
				const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>&,
				const MeshNumbering<Dimension2>&, int n, int num_loc) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
  
  
  void ElementReference<Dimension2, 4>
  ::ComputeLocalProlongation(FiniteElementProjector& proj,
			     Matrix<Real_wp>& LocalProlongation,
			     const ElementReference<Dimension2, 4>& FaceCoarse,
			     const ElementReference<Dimension2, 4>& FaceFine) const
  {
    cout << "Method not implemented for this finite element" << endl;
    abort(); 
  }
      
}

#define MONTJOIE_FILE_FINITE_ELEMENT_HCURL_AXI_CXX
#endif

