#ifndef MONTJOIE_FILE_FINITE_ELEMENT_HCURL_AXI_HXX

namespace Montjoie
{
  
  //! Class for finite element definition for axisymmetric Maxwell equations
  /*!
   axisymmetric Maxwell equations can be solved with H(curl)/H^1 elements
   this class deals how to treat the transformation F_i when using these elements
   */
  template<>
  class ElementReference<Dimension2, 4> : public ElementGeomReferenceContainer<Dimension2>
  {
  public :
    enum{nb_components_u = 3, nb_components_grad = 3};

    typedef ElementReference<Dimension1, 1> ElementReferenceTrace;
    
    IVect ScalarDof, VectorialDof;

  private:
    ElementReference<Dimension2, 1>& fe_h1;
    ElementReference<Dimension2, 2>& fe_hcurl;
    
  protected:
    IVect ScalarPtsDof, VectorialPtsDof;
    IVect ScalarQuadrature, VectorialQuadrature;
    VectR2 points2d_inside;
    int nb_points_quadrature_inside, nb_dof_loc, nb_dof_boundaries;
    Matrix<int> EdgesDof, LocalScalNumber, LocalVecNumber;
    Vector<IVect> ScalarPtsDofFace, VectorialPtsDofFace;
    Vector<IVect> num_quad_points_surf;
    Vector<int> power_two_face, PosDofOnFace;
    Matrix<Real_wp, General, RowMajor> ChE, ChH, ChEteta, ChHteta, RhE, RhEteta;
    TinyVector<Matrix<Real_wp>, 4> RhSurf;

    //! number of points needed to project a function on finite element space
    int nb_points_dof_inside;
    //! list of dof points on each surface
    Vector<IVect> num_dof_points_surf;

    EdgeLobattoReference edge_lob;
    
  public:
    ElementReference(ElementReference<Dimension2, 1>&,
		     ElementReference<Dimension2, 2>&,
		     ElementGeomReference<Dimension2>&);
    
    virtual ~ElementReference<Dimension2, 4>();
    
    // Inline methods
    int GetNbDof() const;
    int GetNbDofBoundaries() const;
    int GetNbPointsQuadratureInside() const;
    int GetNbVectorialDof() const;
    int GetNbScalarDof() const;
    int GetNbDofH() const;
    int GetNbDofHteta() const;
    
    int GetNbDofBoundary(int) const;
    int GetQuadNumber(int num_loc, int j) const;
    int GetLocalNumber(int num_loc, int j) const;
    int GetLocalNumberScal(int num_loc, int j) const;
    int GetLocalNumberVec(int num_loc, int j) const;
    int GetNbVectorialDofOnEdge() const;
    bool IsTangentialDof(int j, int num_loc) const;
    
    const VectR2& PointsQuadInsideND() const;
    
    int GetNbPointsDofInside() const;
    int GetNbPointsDofSurface(int num_loc) const;
    int GetPointDofNumber(int num_loc, int k) const;
    
    bool UseQuadraturePointsForSh() const;
    bool LinearSparseMassMatrix() const;
    bool DiagonalMassMatrix() const;
    bool LumpedMassMatrix() const;
    bool DiscontinuousElement() const;
    void SetPiolaTransform(bool flag = true);
    int GetTypeIntegrationEdge() const;
    int GetOrder() const;

    const ElementReference<Dimension1, 1>& GetSurfaceFiniteElement(int n) const { return edge_lob; }
    
    void FjSurfaceElem(const VectR_N& s, SetPoints<Dimension2>& res,
		       const Mesh<Dimension2>& mesh, int nquad, int ne) const;
    
    void DFjSurfaceElem(const VectR_N& s, const SetPoints<Dimension2>& res,
			SetMatrices<Dimension2>& mat,
			const Mesh<Dimension2>& mesh, int nquad, int ne) const;
    
    void FjSurfaceElemDof(const VectR_N& s, SetPoints<Dimension2>& res,
			  const Mesh<Dimension2>& mesh, int nquad, int ne) const;
    
    void DFjSurfaceElemDof(const VectR_N& s, const SetPoints<Dimension2>& res,
			   SetMatrices<Dimension2>& mat,
			   const Mesh<Dimension2>& mesh, int nquad, int ne) const;
    
    void SetInterpolationElement(Mesh<Dimension2>& mesh, MeshNumbering<Dimension2>&);
    

    void ComputeValuesPhiHtetaRef(const R2& point, VectReal_wp& phi) const;
    
    template<class Vector1, class Vector2>
    void ApplyChTransposeH(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChTransposeE(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChE(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChH(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChTransposeHteta(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChTransposeEteta(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChEteta(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyChHteta(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyRhTransposeEteta(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyRhTransposeE(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyRhEteta(const Vector1& U, Vector2& V) const;
    
    template<class Vector1, class Vector2>
    void ApplyRhE(const Vector1& U, Vector2& V) const;

    template<class Vector1, class Vector2>
    void ApplyRhTangential(const Vector1& U, Vector2& V, int num_loc) const;

    bool OptimizedComputationMassMatrix() const;
    bool SparseMassMatrix() const;
    
    // Other methods
    void ApplyRhQuadratureTranspose(const VectReal_wp&, VectReal_wp&) const;
    void ApplyRhTranspose(const VectReal_wp&, VectReal_wp&) const;
    void ApplyChTranspose(const VectReal_wp&, VectReal_wp&) const;
    void ApplyCh(const VectReal_wp&, VectReal_wp&) const;

    void MltMassMatrix(VectReal_wp&) const;
    void SolveMassMatrix(VectReal_wp&) const;
    void SolveCholesky(const SeldonTranspose&, VectReal_wp&) const;
    
    void ComputeMassMatrix(VirtualMatrix<Real_wp>&, const VectReal_wp&) const;
    void IntegrateMassMatrix(const VirtualMatrix<Real_wp>&, const VectReal_wp&) const;
    
    template<class T>
    void ComputeGaussIntegralSurface(const Vector<Matrix2_2>& dfjm1,
                                     const VectReal_wp& weights, const VectReal_wp& ds,
                                     Vector<Vector<T> > & feval, Vector<Vector<T> >& res,
                                     const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>&,
				     const MeshNumbering<Dimension2>&, int n, int num_loc) const;

    inline void ComputeProjectionPointsSurf(int num_loc, const VectReal_wp& ptsA,
					    const VectReal_wp& ptsB, Matrix<Real_wp>& proj) const {}
    
    size_t GetMemorySize() const;
    
    virtual void ConstructNumberMap(NumberMap&, int dg) const = 0;
    
    // construction of finite element
    void ConstructFiniteElement(int r, int rquad = 0, int rgeom = 0, int type_quad = -1,
				int rsurf = 0, int type_surf = -1);

    void ConstructFunctions();
    void ConstructMassMatrix();
    
    static void MergePoints(const VectR2&, const VectR2&, VectR2&, IVect&, IVect&);
    
    void ComputeValuesPhiHRef(const R2& point, VectR2& phi) const;
    
    void ComputeValuesPhi(const R2& point_loc, const R2& point_glob,
			  Vector<R3>& res, const Matrix2_2& dfjm1,
                          const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
			  const MeshNumbering<Dimension2>& mesh_num, int nquad) const;
    
    void ComputeValuesGradientPhi(const R2& point_loc, const R2& point_glob,
				  Vector<R3>& res, const Matrix2_2& dfjm1,
                                  const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>&,
				  const MeshNumbering<Dimension2>& mesh_num, int nquad) const;
    
    template<class T>
    static void GetCurlFromGradient(const Vector<Vector<T> >& dU,
				    Vector<Vector<T> >& curlU);
    
    void ComputeValuesPhiBoundary(Matrix<R3>& ValuePhi_Boundary,
                                  const SetPoints<Dimension2>& PointsElem,
                                  const SetMatrices<Dimension2>& MatricesElem,
                                  const ElementReference<Dimension2, 4>& FaceBasis,
				  const Mesh<Dimension2>& mesh,
				  const MeshNumbering<Dimension2>& mesh_num,
				  int nquad, int num_loc) const;

    void ComputeValuesGradientPhiBoundary(Matrix<R3>& GradientPhi_Boundary,
                                          const SetPoints<Dimension2>& PointsElem,
                                          const SetMatrices<Dimension2>& MatricesElem,
                                          const ElementReference<Dimension2, 4>& FaceBasis,
                                          const Mesh<Dimension2>&,
					  const MeshNumbering<Dimension2>& mesh_num,
					  int nquad, int num_loc) const;
    
    template<class T>
    void ComputeValueNodalBoundary(const Vector<T>& Uloc, Vector<T>& Uboundary,
				   const ElementReference<Dimension2, 4>& FaceBasis,
                                   const Mesh<Dimension2>& mesh, int n, int num_loc) const;

    template<class T>
    void ComputeQuadratureValues(const SetPoints<Dimension2>& PointsElem,
				 const SetMatrices<Dimension2>& MatricesElem,
				 const Vector<Vector<T> >& Uloc, Vector<Vector<T> >& Uloc_node,
				 Vector<Vector<T> >& dUloc_node, bool compute_u, bool compute_curl,
				 const ElementReference<Dimension2, 4>& FaceBasis,
				 const Mesh<Dimension2>& mesh, int iquad) const;

    template<class T>
    void ComputeNodalValues(const SetPoints<Dimension2>& PointsElem,
			    const SetMatrices<Dimension2>& MatricesElem,
			    const Vector<Vector<T> >& Uloc, Vector<Vector<T> >& Uloc_node,
			    const ElementReference<Dimension2, 4>& FaceBasis,
			    const Mesh<Dimension2>& mesh, int i) const;
    
    template<class T>
    void ComputeValueBoundary(const SetPoints<Dimension2>& pts,
			      const SetMatrices<Dimension2>& mat,
			      const Vector<Vector<T> >& feval, Vector<Vector<TinyVector<T, 3> > >& res,
			      const ElementReference<Dimension2, 4>& FaceBasis,
			      const Mesh<Dimension2>& mesh, int n, int num_loc) const;
    
    template<class T>
    void ComputeGradientBoundary(const SetPoints<Dimension2>& pts,
				 const SetMatrices<Dimension2>& mat,
				 const Vector<Vector<T> >& feval, Vector<Vector<TinyVector<T, 3> > >& res,
				 const ElementReference<Dimension2, 4>& FaceBasis,
				 const Mesh<Dimension2>& mesh, int n, int num_loc) const;
    
    template<class T>
    void ComputeProjectionDof(const SetPoints<Dimension2>& PointsElem,
			      const SetMatrices<Dimension2>& MatricesElem,
			      Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
                              const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
			      const MeshNumbering<Dimension2>& mesh_num, int n) const;

    template<class T>
    void ComputeProjectionSurfaceDof(const SetPoints<Dimension2>& PointsElem,
				     const SetMatrices<Dimension2>& MatricesElem,
				     Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
				     const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
				     const MeshNumbering<Dimension2>& mesh_num, int n, int num_loc) const;
    
    template<class T>
    void ComputeIntegral(const SetPoints<Dimension2>& pts,
			 const SetMatrices<Dimension2>& mat,
			 Vector<Vector<T> >& feval, Vector<Vector<T> > & res,
			 const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
			 const MeshNumbering<Dimension2>& mesh_num, int nquad) const;
    
    template<class T>
    void ComputeIntegralH(const SetPoints<Dimension2>& pts,
                          const SetMatrices<Dimension2>& mat,
                          Vector<T>& feval, Vector<T>& contribH,
                          Vector<T>& contribHteta, const ElementReference<Dimension2, 4>& FaceBasis,
                          const Mesh<Dimension2>& mesh, int nquad) const;
    
    template<class T>
    void ComputeIntegralGradient(const SetPoints<Dimension2>& pts,
				 const SetMatrices<Dimension2>& mat,
				 Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
				 const ElementReference<Dimension2, 4>& FaceBasis, const Mesh<Dimension2>& mesh,
				 const MeshNumbering<Dimension2>& mesh_num,  int nquad) const;
    
    template<class T>
    void ComputeIntegralSurface(const SetPoints<Dimension2>& pts,
				const SetMatrices<Dimension2>& mat,
				Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
				const ElementReference<Dimension2, 4>& FaceBasis,
				const Mesh<Dimension2>& mesh,
				const MeshNumbering<Dimension2>& mesh_num, 
				int n, int num_loc) const;

    template<class T>
    void ComputeIntegralSurfaceHDG(const SetPoints<Dimension2>& pts,
				   const SetMatrices<Dimension2>& mat,
				   Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
				   const ElementReference<Dimension2, 4>& Fb_vol,  
                                   const ElementReferenceTrace& FaceBasis,
				   const Mesh<Dimension2>& mesh,
				   const MeshNumbering<Dimension2>& mesh_num, 
				   int n, int num_loc) const {}

    template<class T>
    void ComputeIntegralSurfaceGradient(const SetPoints<Dimension2>& pts,
					const SetMatrices<Dimension2>& mat,
					Vector<Vector<T> >& feval, Vector<Vector<T> >& res,
					const ElementReference<Dimension2, 4>& FaceBasis,
					const Mesh<Dimension2>& mesh,
					const MeshNumbering<Dimension2>& mesh_num,
					int n, int num_loc) const;
    
    void ComputeLocalProlongation(FiniteElementProjector& proj,
				  Matrix<Real_wp>& LocalProlongation,
				  const ElementReference<Dimension2, 4>& FaceCoarse,
				  const ElementReference<Dimension2, 4>& FaceFine) const;
    
    // inline static methods
    template<class T0>
    static void ApplyRotationCyclic(int n, FftInterface<Complex_wp>& fft_interface,
					   TinyVector<T0, 3>& val_u);
    
    template<class T0>
    static void ApplyInverseRotationCyclic(int n, FftInterface<Complex_wp>& fft_interface,
					   TinyVector<T0, 3>& val_u);
    
    template<class T>
    static void ApplyInverseRotationCyclic(int n, FftInterface<Complex_wp>& fft_interface,
                                           Vector<T>& val_u);

    inline void PickNearDofs(int pos, const VectBool& DofUsed, IVect& ListeDof, int nb_dof) const {}

    inline void ComputeLocalProlongationLowOrder(FiniteElementProjector& proj,
						 Matrix<Real_wp>& LocalProlongation,
						 const ElementReference<Dimension2, 4>& FaceCoarse) const {}
        
  };

  
  template<class TypeEltH1, class TypeEltHcurl>
  class FiniteElementHcurlAxi : public ElementReference<Dimension2, 4>
  {
  protected:
    TypeEltH1 finite_element_h1;
    TypeEltHcurl finite_element_hcurl;

  public:
    FiniteElementHcurlAxi();
    
  };
  
}

#define MONTJOIE_FILE_FINITE_ELEMENT_HCURL_AXI_HXX
#endif
