#ifndef MONTJOIE_FILE_TETRAHEDRON_QUADRATURE_CXX

namespace Montjoie
{
  
  //! constructs quadrature formulas over unit tetrahedron
  /*!
    \param[in] p polynomials of degree at most p must be integrated exactly
    \param[out] points3d 3-D integration points \xi_i
    \param[out] weights3d 3-D integration weights \omega_i
    \param[in] type_quad type of quadrature
    The quadrature formulas can be used to integrate :
    \int_K f(x, y, z) = \sum \omega_i f(\xi_i)
    where K is the unit tetrahedron defined by :
    x, y, z >= 0  and x+y+z <= 1
   */
  void TetrahedronQuadrature::
  ConstructQuadrature(int p, VectR3& points3d, VectReal_wp& weights3d, int type_quad)
  {
    // quadrature formula over a tetrahedron
    // Gauss-like quadrature -> points are strictely inside the tetrahedron
    	
    bool formula_not_available = false;
    int n1 = 0, n4 = 0, n6 = 0, n12 = 0, n24 = 0;
    int nb_points = 0;
    
    if (type_quad == QUADRATURE_GAUSS)
      {
        switch (p)
          {
          case 0:
          case 1 : 
            { 
              n1 = 1; n4 = 0; n6 = 0; n12 = 0; n24 = 0;
              nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
              points3d.Reallocate(nb_points); 
              weights3d.Reallocate(nb_points); 
              points3d(0).Init(0.25, 0.25, 0.25);
              weights3d(0) = 0.1666666666666666;
            } 
            break; 
          case 2 : 
            { 
              n1 = 0; n4 = 1; n6 = 0; n12 = 0; n24 = 0;
              nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
              points3d.Reallocate(nb_points); 
              weights3d.Reallocate(nb_points); 
              points3d(0).Init(0.1381966011250105, 0.1381966011250105, 0.5854101966249685);
              weights3d(0) = 0.04166666666666662;
            } 
            break; 
          case 3 : 
            { 
              n1 = 1; n4 = 1; n6 = 0; n12 = 0; n24 = 0;
              nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
              points3d.Reallocate(nb_points); 
              weights3d.Reallocate(nb_points); 
              points3d(0).Init(0.25, 0.25, 0.25);
              weights3d(0) = -0.1333333333333334;
              points3d(1).Init(0.1666666666666665, 0.1666666666666665, 0.5000000000000004);
              weights3d(1) = 0.075;
            } 
            break; 
          case 4 : 
            { 
              n1 = 1; n4 = 1; n6 = 1; n12 = 0; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = -0.0131555555555555;
	      points3d(1).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(1) = 0.00762222222222225;
	      points3d(5).Init(0.399403576166799, 0.399403576166799, 0.100596423833201);
	      weights3d(5) = 0.02488888888888887;
	    } 
	    break; 
	  case 5 : 
	    { 
	      n1 = 0; n4 = 2; n6 = 1; n12 = 0; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.09273525031089103, 0.09273525031089103, 0.7217942490673269);
	      weights3d(0) = 0.01224884051939363;
	      points3d(4).Init(0.3108859192633005, 0.3108859192633005, 0.06734224221009849);
	      weights3d(4) = 0.01878132095300263;
	      points3d(8).Init(0.4544962958743505, 0.4544962958743505, 0.04550370412564952);
	      weights3d(8) = 0.007091003462846875;
	    } 
	    break; 
	  case 6 : 
	    { 
	      n1 = 0; n4 = 3; n6 = 0; n12 = 1; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.214602871259152, 0.214602871259152, 0.356191386222544);
	      weights3d(0) = 0.006653791709694625;
	      points3d(4).Init(0.04067395853461148, 0.04067395853461148, 0.8779781243961655);
	      weights3d(4) = 0.00167953517588675;
	      points3d(8).Init(0.3223378901422755, 0.3223378901422755, 0.03298632957317338);
	      weights3d(8) = 0.0092261969239425;
	      points3d(12).Init(0.0636610018750175, 0.0636610018750175, 0.269672331458316);
	      weights3d(12) = 0.00803571428571425;
	    } 
	    break; 
	  case 7 : 
	    { 
	      n1 = 1; n4 = 3; n6 = 1; n12 = 1; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = 0.01826422346610887;
	      points3d(1).Init(0.07821319233031798, 0.07821319233031798, 0.7653604230090461);
	      weights3d(1) = 0.01059994152441363;
	      points3d(5).Init(0.121843216663905, 0.121843216663905, 0.6344703500082849);
	      weights3d(5) = -0.06251774011433188;
	      points3d(9).Init(0.3325391644464205, 0.3325391644464205, 0.00238250666073847);
	      weights3d(9) = 0.0048914252630735;
	      points3d(13).Init(0.5, 0.5, 0);
	      weights3d(13) = 0.00097001763668425;
	      points3d(19).Init(0.09999999999999998, 0.09999999999999998, 0.2);
	      weights3d(19) = 0.02755731922398588;
	    } 
	    break; 
	  case 8 : 
	    { 
	      n1 = 1; n4 = 3; n6 = 1; n12 = 2; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = -0.02050018865863987;
	      points3d(1).Init(0.206829931610673, 0.206829931610673, 0.3795102051679811);
	      weights3d(1) = 0.01425030582286687;
	      points3d(5).Init(0.08210358831054648, 0.08210358831054648, 0.7536892350683606);
	      weights3d(5) = 0.001967033313133875;
	      points3d(9).Init(0.005781950505197975, 0.005781950505197975, 0.982654148484406);
	      weights3d(9) = 0.000169834109092875;
	      points3d(13).Init(0.05053274001889402, 0.05053274001889402, 0.449467259981106);
	      weights3d(13) = 0.00457968382446725;
	      points3d(19).Init(0.229066536116811, 0.229066536116811, 0.03563958278853402);
	      weights3d(19) = 0.005704485808681875;
	      points3d(31).Init(0.03660774955319751, 0.03660774955319751, 0.1904860419346335);
	      weights3d(31) = 0.002140519141162125;
	    } 
	    break; 
	  case 9 : 
	    { 
	      n1 = 1; n4 = 4; n6 = 0; n12 = 3; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = -0.1377990383261086;
	      points3d(1).Init(0.04835103854973699, 0.04835103854973699, 0.854946884350789);
	      weights3d(1) = 0.00186533656908525;
	      points3d(5).Init(0.3245792801178825, 0.3245792801178825, 0.02626215964635248);
	      weights3d(5) = 0.004309423969493375;
	      points3d(9).Init(0.114616540223995, 0.114616540223995, 0.656150379328015);
	      weights3d(9) = -0.09018476648120149;
	      points3d(13).Init(0.225489951911514, 0.225489951911514, 0.323530144265458);
	      weights3d(13) = 0.0446725762025115;
	      points3d(17).Init(0.13162780924687, 0.13162780924687, 0.08366470161718498);
	      weights3d(17) = 0.03470040588455075;
	      points3d(29).Init(0.433951461411407, 0.433951461411407, 0.1077698595494285);
	      weights3d(29) = 0.003352583902660625;
	      points3d(41).Init(-0.001376277318138053, -0.001376277318138053, 0.2765534726368075);
	      weights3d(41) = 0.000431628875557;
	    } 
	    break; 
	  case 11 : 
	    { 
	      n1 = 1; n4 = 7; n6 = 2; n12 = 7; n24 = 0;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = -0.116617538839761;
	      points3d(1).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(1) = 0.045362824065081;
	      points3d(5).Init(0.2142857142857145, 0.2142857142857145, 0.357142857142857);
	      weights3d(5) = 0.045362824065081;
	      points3d(9).Init(0.08333333333333348, 0.08333333333333348, 0.75);
	      weights3d(9) = -0.1165234765234765;
	      points3d(13).Init(0.09999999999999998, 0.09999999999999998, 0.7);
	      weights3d(13) = 0.1019372899798247;
	      points3d(17).Init(0.3, 0.3, 0.1);
	      weights3d(17) = 0.1019372899798247;
	      points3d(21).Init(0.125, 0.125, 0.625);
	      weights3d(21) = -0.03502538613649725;
	      points3d(25).Init(0.1666666666666665, 0.1666666666666665, 0.5);
	      weights3d(25) = 0.004068080357142875;
	      points3d(29).Init(0.08333333333333348, 0.08333333333333348, 0.4166666666666665);
	      weights3d(29) = -0.1165234765234765;
	      points3d(35).Init(0.125, 0.125, 0.375);
	      weights3d(35) = -0.03502538613649725;
	      points3d(41).Init(0.07142857142857151, 0.07142857142857151, 0.642857142857143);
	      weights3d(41) = 0.045362824065081;
	      points3d(53).Init(0.07142857142857151, 0.07142857142857151, 0.5);
	      weights3d(53) = 0.045362824065081;
	      points3d(65).Init(0.2142857142857145, 0.2142857142857145, 0.5);
	      weights3d(65) = 0.045362824065081;
	      points3d(77).Init(0.357142857142857, 0.357142857142857, 0.2142857142857145);
	      weights3d(77) = 0.045362824065081;
	      points3d(89).Init(0.08333333333333348, 0.08333333333333348, 0.5833333333333335);
	      weights3d(89) = -0.1165234765234765;
	      points3d(101).Init(0.25, 0.25, 0.4166666666666665);
	      weights3d(101) = -0.1165234765234765;
	      points3d(113).Init(0.09999999999999998, 0.09999999999999998, 0.5);
	      weights3d(113) = 0.1019372899798247;
	    } 
	    break; 
	  case 13 : 
	    { 
	      n1 = 1; n4 = 10; n6 = 4; n12 = 10; n24 = 1;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = 0.1498221692084126;
	      points3d(1).Init(0.0625, 0.0625, 0.8125);
	      weights3d(1) = 0.0525509090165175;
	      points3d(5).Init(0.3125, 0.3125, 0.0625);
	      weights3d(5) = 0.0525509090165175;
	      points3d(9).Init(0.1875, 0.1875, 0.4375);
	      weights3d(9) = 0.0525509090165175;
	      points3d(13).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(13) = -0.1481852252792646;
	      points3d(17).Init(0.2142857142857145, 0.2142857142857145, 0.357142857142857);
	      weights3d(17) = -0.1481852252792646;
	      points3d(21).Init(0.08333333333333348, 0.08333333333333348, 0.75);
	      weights3d(21) = 0.149815898387327;
	      points3d(25).Init(0.09999999999999998, 0.09999999999999998, 0.7);
	      weights3d(25) = -0.06534441665373375;
	      points3d(29).Init(0.3, 0.3, 0.1);
	      weights3d(29) = -0.06534441665373375;
	      points3d(33).Init(0.125, 0.125, 0.625);
	      weights3d(33) = 0.01167512871216575;
	      points3d(37).Init(0.1666666666666665, 0.1666666666666665, 0.5);
	      weights3d(37) = -0.000665685876623375;
	      points3d(41).Init(0.0625, 0.0625, 0.4375);
	      weights3d(41) = 0.0525509090165175;
	      points3d(47).Init(0.1875, 0.1875, 0.3125);
	      weights3d(47) = 0.0525509090165175;
	      points3d(53).Init(0.08333333333333348, 0.08333333333333348, 0.4166666666666665);
	      weights3d(53) = 0.149815898387327;
	      points3d(59).Init(0.125, 0.125, 0.375);
	      weights3d(59) = 0.01167512871216575;
	      points3d(65).Init(0.0625, 0.0625, 0.6875);
	      weights3d(65) = 0.0525509090165175;
	      points3d(77).Init(0.0625, 0.0625, 0.5625);
	      weights3d(77) = 0.0525509090165175;
	      points3d(89).Init(0.1875, 0.1875, 0.5625);
	      weights3d(89) = 0.0525509090165175;
	      points3d(101).Init(0.07142857142857151, 0.07142857142857151, 0.642857142857143);
	      weights3d(101) = -0.1481852252792646;
	      points3d(113).Init(0.07142857142857151, 0.07142857142857151, 0.5);
	      weights3d(113) = -0.1481852252792646;
	      points3d(125).Init(0.2142857142857145, 0.2142857142857145, 0.5);
	      weights3d(125) = -0.1481852252792646;
	      points3d(137).Init(0.357142857142857, 0.357142857142857, 0.2142857142857145);
	      weights3d(137) = -0.1481852252792646;
	      points3d(149).Init(0.08333333333333348, 0.08333333333333348, 0.5833333333333335);
	      weights3d(149) = 0.149815898387327;
	      points3d(161).Init(0.25, 0.25, 0.4166666666666665);
	      weights3d(161) = 0.149815898387327;
	      points3d(173).Init(0.09999999999999998, 0.09999999999999998, 0.5);
	      weights3d(173) = -0.06534441665373375;
	      points3d(185).Init(0.1875, 0.3125, 0.4375);
	      weights3d(185) = 0.0525509090165175;
	    } 
	    break; 
	  case 15 : 
	    { 
	      n1 = 1; n4 = 12; n6 = 4; n12 = 17; n24 = 2;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = -0.1198530444668011;
	      points3d(1).Init(0.05555555555555552, 0.05555555555555552, 0.8333333333333335);
	      weights3d(1) = 0.064317124103657;
	      points3d(5).Init(0.1666666666666665, 0.1666666666666665, 0.5);
	      weights3d(5) = 0.06440033483823487;
	      points3d(9).Init(0.277777777777778, 0.277777777777778, 0.166666666666666);
	      weights3d(9) = 0.064317124103657;
	      points3d(13).Init(0.0625, 0.0625, 0.8125);
	      weights3d(13) = -0.1978387162974776;
	      points3d(17).Init(0.3125, 0.3125, 0.0625);
	      weights3d(17) = -0.1978387162974776;
	      points3d(21).Init(0.1875, 0.1875, 0.4375);
	      weights3d(21) = -0.1978387162974776;
	      points3d(25).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(25) = 0.226908626208874;
	      points3d(29).Init(0.2142857142857145, 0.2142857142857145, 0.357142857142857);
	      weights3d(29) = 0.226908626208874;
	      points3d(33).Init(0.08333333333333348, 0.08333333333333348, 0.75);
	      weights3d(33) = -0.1198527187098616;
	      points3d(37).Init(0.09999999999999998, 0.09999999999999998, 0.7);
	      weights3d(37) = 0.02917161457755975;
	      points3d(41).Init(0.3, 0.3, 0.1);
	      weights3d(41) = 0.02917161457755975;
	      points3d(45).Init(0.125, 0.125, 0.625);
	      weights3d(45) = -0.00287387783684075;
	      points3d(49).Init(0.0625, 0.0625, 0.4375);
	      weights3d(49) = -0.1978387162974776;
	      points3d(55).Init(0.1875, 0.1875, 0.3125);
	      weights3d(55) = -0.1978387162974776;
	      points3d(61).Init(0.08333333333333348, 0.08333333333333348, 0.4166666666666665);
	      weights3d(61) = -0.1198527187098616;
	      points3d(67).Init(0.125, 0.125, 0.375);
	      weights3d(67) = -0.00287387783684075;
	      points3d(73).Init(0.05555555555555552, 0.05555555555555552, 0.722222222222222);
	      weights3d(73) = 0.064317124103657;
	      points3d(85).Init(0.05555555555555552, 0.05555555555555552, 0.6111111111111109);
	      weights3d(85) = 0.064317124103657;
	      points3d(97).Init(0.05555555555555552, 0.05555555555555552, 0.5);
	      weights3d(97) = 0.064317124103657;
	      points3d(109).Init(0.1666666666666665, 0.1666666666666665, 0.6111111111111109);
	      weights3d(109) = 0.064317124103657;
	      points3d(121).Init(0.388888888888889, 0.388888888888889, 0.1666666666666665);
	      weights3d(121) = 0.064317124103657;
	      points3d(133).Init(0.277777777777778, 0.277777777777778, 0.388888888888889);
	      weights3d(133) = 0.064317124103657;
	      points3d(145).Init(0.1666666666666665, 0.1666666666666665, 0.388888888888889);
	      weights3d(145) = 0.064317124103657;
	      points3d(157).Init(0.0625, 0.0625, 0.6875);
	      weights3d(157) = -0.1978387162974776;
	      points3d(169).Init(0.0625, 0.0625, 0.5625);
	      weights3d(169) = -0.1978387162974776;
	      points3d(181).Init(0.1875, 0.1875, 0.5625);
	      weights3d(181) = -0.1978387162974776;
	      points3d(193).Init(0.07142857142857151, 0.07142857142857151, 0.642857142857143);
	      weights3d(193) = 0.226908626208874;
	      points3d(205).Init(0.07142857142857151, 0.07142857142857151, 0.5);
	      weights3d(205) = 0.226908626208874;
	      points3d(217).Init(0.2142857142857145, 0.2142857142857145, 0.5);
	      weights3d(217) = 0.226908626208874;
	      points3d(229).Init(0.357142857142857, 0.357142857142857, 0.2142857142857145);
	      weights3d(229) = 0.226908626208874;
	      points3d(241).Init(0.08333333333333348, 0.08333333333333348, 0.5833333333333335);
	      weights3d(241) = -0.1198527187098616;
	      points3d(253).Init(0.25, 0.25, 0.4166666666666665);
	      weights3d(253) = -0.1198527187098616;
	      points3d(265).Init(0.09999999999999998, 0.09999999999999998, 0.5);
	      weights3d(265) = 0.02917161457755975;
	      points3d(277).Init(0.1666666666666665, 0.277777777777778, 0.5);
	      weights3d(277) = 0.064317124103657;
	      points3d(301).Init(0.1875, 0.3125, 0.4375);
	      weights3d(301) = -0.1978387162974776;
	    } 
	    break; 
	  case 17 : 
	    { 
	      n1 = 1; n4 = 14; n6 = 6; n12 = 25; n24 = 4;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = 0.1496235203137462;
	      points3d(1).Init(0.04999999999999999, 0.04999999999999999, 0.85);
	      weights3d(1) = 0.08220635246624325;
	      points3d(5).Init(0.15, 0.15, 0.55);
	      weights3d(5) = 0.08220635246624325;
	      points3d(9).Init(0.05555555555555552, 0.05555555555555552, 0.8333333333333335);
	      weights3d(9) = -0.2741940553892745;
	      points3d(13).Init(0.1666666666666665, 0.1666666666666665, 0.5);
	      weights3d(13) = -0.2742022850223646;
	      points3d(17).Init(0.277777777777778, 0.277777777777778, 0.166666666666666);
	      weights3d(17) = -0.2741940553892745;
	      points3d(21).Init(0.0625, 0.0625, 0.8125);
	      weights3d(21) = 0.3517132734177379;
	      points3d(25).Init(0.3125, 0.3125, 0.0625);
	      weights3d(25) = 0.3517132734177379;
	      points3d(29).Init(0.1875, 0.1875, 0.4375);
	      weights3d(29) = 0.3517132734177379;
	      points3d(33).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(33) = -0.2180102487104867;
	      points3d(37).Init(0.2142857142857145, 0.2142857142857145, 0.357142857142857);
	      weights3d(37) = -0.2180102487104867;
	      points3d(41).Init(0.08333333333333348, 0.08333333333333348, 0.75);
	      weights3d(41) = 0.06741715427429712;
	      points3d(45).Init(0.09999999999999998, 0.09999999999999998, 0.7);
	      weights3d(45) = -0.00972387152585325;
	      points3d(49).Init(0.3, 0.3, 0.1);
	      weights3d(49) = -0.00972387152585325;
	      points3d(53).Init(0.125, 0.125, 0.625);
	      weights3d(53) = 0.000547405302255375;
	      points3d(57).Init(0.04999999999999999, 0.04999999999999999, 0.45);
	      weights3d(57) = 0.08220635246624325;
	      points3d(63).Init(0.15, 0.15, 0.35);
	      weights3d(63) = 0.08220635246624325;
	      points3d(69).Init(0.0625, 0.0625, 0.4375);
	      weights3d(69) = 0.3517132734177379;
	      points3d(75).Init(0.1875, 0.1875, 0.3125);
	      weights3d(75) = 0.3517132734177379;
	      points3d(81).Init(0.08333333333333348, 0.08333333333333348, 0.4166666666666665);
	      weights3d(81) = 0.06741715427429712;
	      points3d(87).Init(0.125, 0.125, 0.375);
	      weights3d(87) = 0.000547405302255375;
	      points3d(93).Init(0.04999999999999999, 0.04999999999999999, 0.75);
	      weights3d(93) = 0.08220635246624325;
	      points3d(105).Init(0.04999999999999999, 0.04999999999999999, 0.65);
	      weights3d(105) = 0.08220635246624325;
	      points3d(117).Init(0.04999999999999999, 0.04999999999999999, 0.55);
	      weights3d(117) = 0.08220635246624325;
	      points3d(129).Init(0.15, 0.15, 0.65);
	      weights3d(129) = 0.08220635246624325;
	      points3d(141).Init(0.25, 0.25, 0.45);
	      weights3d(141) = 0.08220635246624325;
	      points3d(153).Init(0.35, 0.35, 0.25);
	      weights3d(153) = 0.08220635246624325;
	      points3d(165).Init(0.15, 0.15, 0.45);
	      weights3d(165) = 0.08220635246624325;
	      points3d(177).Init(0.25, 0.25, 0.35);
	      weights3d(177) = 0.08220635246624325;
	      points3d(189).Init(0.05555555555555552, 0.05555555555555552, 0.722222222222222);
	      weights3d(189) = -0.2741940553892745;
	      points3d(201).Init(0.05555555555555552, 0.05555555555555552, 0.6111111111111109);
	      weights3d(201) = -0.2741940553892745;
	      points3d(213).Init(0.05555555555555552, 0.05555555555555552, 0.5);
	      weights3d(213) = -0.2741940553892745;
	      points3d(225).Init(0.1666666666666665, 0.1666666666666665, 0.6111111111111109);
	      weights3d(225) = -0.2741940553892745;
	      points3d(237).Init(0.388888888888889, 0.388888888888889, 0.1666666666666665);
	      weights3d(237) = -0.2741940553892745;
	      points3d(249).Init(0.277777777777778, 0.277777777777778, 0.388888888888889);
	      weights3d(249) = -0.2741940553892745;
	      points3d(261).Init(0.1666666666666665, 0.1666666666666665, 0.388888888888889);
	      weights3d(261) = -0.2741940553892745;
	      points3d(273).Init(0.0625, 0.0625, 0.6875);
	      weights3d(273) = 0.3517132734177379;
	      points3d(285).Init(0.0625, 0.0625, 0.5625);
	      weights3d(285) = 0.3517132734177379;
	      points3d(297).Init(0.1875, 0.1875, 0.5625);
	      weights3d(297) = 0.3517132734177379;
	      points3d(309).Init(0.07142857142857151, 0.07142857142857151, 0.642857142857143);
	      weights3d(309) = -0.2180102487104867;
	      points3d(321).Init(0.07142857142857151, 0.07142857142857151, 0.5);
	      weights3d(321) = -0.2180102487104867;
	      points3d(333).Init(0.2142857142857145, 0.2142857142857145, 0.5);
	      weights3d(333) = -0.2180102487104867;
	      points3d(345).Init(0.357142857142857, 0.357142857142857, 0.2142857142857145);
	      weights3d(345) = -0.2180102487104867;
	      points3d(357).Init(0.08333333333333348, 0.08333333333333348, 0.5833333333333335);
	      weights3d(357) = 0.06741715427429712;
	      points3d(369).Init(0.25, 0.25, 0.4166666666666665);
	      weights3d(369) = 0.06741715427429712;
	      points3d(381).Init(0.09999999999999998, 0.09999999999999998, 0.5);
	      weights3d(381) = -0.00972387152585325;
	      points3d(393).Init(0.15, 0.25, 0.55);
	      weights3d(393) = 0.08220635246624325;
	      points3d(417).Init(0.15, 0.35, 0.45);
	      weights3d(417) = 0.08220635246624325;
	      points3d(441).Init(0.1666666666666665, 0.277777777777778, 0.5);
	      weights3d(441) = -0.2741940553892745;
	      points3d(465).Init(0.1875, 0.3125, 0.4375);
	      weights3d(465) = 0.3517132734177379;
	    } 
	    break; 
	  case 19 : 
	    { 
	      n1 = 1; n4 = 18; n6 = 6; n12 = 36; n24 = 7;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = -0.42001196942451;
	      points3d(1).Init(0.04545454545454553, 0.04545454545454553, 0.8636363636363635);
	      weights3d(1) = 0.1088239338943121;
	      points3d(5).Init(0.318181818181818, 0.318181818181818, 0.04545454545454597);
	      weights3d(5) = 0.1088239338943121;
	      points3d(9).Init(0.1363636363636365, 0.1363636363636365, 0.5909090909090911);
	      weights3d(9) = 0.1088239338943121;
	      points3d(13).Init(0.2272727272727275, 0.2272727272727275, 0.318181818181818);
	      weights3d(13) = 0.1088239338943121;
	      points3d(17).Init(0.04999999999999999, 0.04999999999999999, 0.85);
	      weights3d(17) = -0.3914588212678253;
	      points3d(21).Init(0.15, 0.15, 0.55);
	      weights3d(21) = -0.3914588212678253;
	      points3d(25).Init(0.05555555555555552, 0.05555555555555552, 0.8333333333333335);
	      weights3d(25) = 0.5552429621632808;
	      points3d(29).Init(0.1666666666666665, 0.1666666666666665, 0.5);
	      weights3d(29) = 0.5552436234730828;
	      points3d(33).Init(0.277777777777778, 0.277777777777778, 0.166666666666666);
	      weights3d(33) = 0.5552429621632808;
	      points3d(37).Init(0.0625, 0.0625, 0.8125);
	      weights3d(37) = -0.3949061315567584;
	      points3d(41).Init(0.3125, 0.3125, 0.0625);
	      weights3d(41) = -0.3949061315567584;
	      points3d(45).Init(0.1875, 0.1875, 0.4375);
	      weights3d(45) = -0.3949061315567584;
	      points3d(49).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(49) = 0.1483680859279701;
	      points3d(53).Init(0.2142857142857145, 0.2142857142857145, 0.357142857142857);
	      weights3d(53) = 0.1483680859279701;
	      points3d(57).Init(0.08333333333333348, 0.08333333333333348, 0.75);
	      weights3d(57) = -0.0285531476926435;
	      points3d(61).Init(0.09999999999999998, 0.09999999999999998, 0.7);
	      weights3d(61) = 0.002532258209857625;
	      points3d(65).Init(0.3, 0.3, 0.1);
	      weights3d(65) = 0.002532258209857625;
	      points3d(69).Init(0.125, 0.125, 0.625);
	      weights3d(69) = -8.3414141296e-05;
	      points3d(73).Init(0.04999999999999999, 0.04999999999999999, 0.45);
	      weights3d(73) = -0.3914588212678253;
	      points3d(79).Init(0.15, 0.15, 0.35);
	      weights3d(79) = -0.3914588212678253;
	      points3d(85).Init(0.0625, 0.0625, 0.4375);
	      weights3d(85) = -0.3949061315567584;
	      points3d(91).Init(0.1875, 0.1875, 0.3125);
	      weights3d(91) = -0.3949061315567584;
	      points3d(97).Init(0.08333333333333348, 0.08333333333333348, 0.4166666666666665);
	      weights3d(97) = -0.0285531476926435;
	      points3d(103).Init(0.125, 0.125, 0.375);
	      weights3d(103) = -8.3414141296e-05;
	      points3d(109).Init(0.04545454545454553, 0.04545454545454553, 0.7727272727272725);
	      weights3d(109) = 0.1088239338943121;
	      points3d(121).Init(0.04545454545454553, 0.04545454545454553, 0.681818181818182);
	      weights3d(121) = 0.1088239338943121;
	      points3d(133).Init(0.04545454545454553, 0.04545454545454553, 0.5909090909090911);
	      weights3d(133) = 0.1088239338943121;
	      points3d(145).Init(0.04545454545454553, 0.04545454545454553, 0.5);
	      weights3d(145) = 0.1088239338943121;
	      points3d(157).Init(0.1363636363636365, 0.1363636363636365, 0.681818181818182);
	      weights3d(157) = 0.1088239338943121;
	      points3d(169).Init(0.409090909090909, 0.409090909090909, 0.1363636363636365);
	      weights3d(169) = 0.1088239338943121;
	      points3d(181).Init(0.2272727272727275, 0.2272727272727275, 0.5);
	      weights3d(181) = 0.1088239338943121;
	      points3d(193).Init(0.1363636363636365, 0.1363636363636365, 0.5);
	      weights3d(193) = 0.1088239338943121;
	      points3d(205).Init(0.1363636363636365, 0.1363636363636365, 0.409090909090909);
	      weights3d(205) = 0.1088239338943121;
	      points3d(217).Init(0.2272727272727275, 0.2272727272727275, 0.409090909090909);
	      weights3d(217) = 0.1088239338943121;
	      points3d(229).Init(0.318181818181818, 0.318181818181818, 0.2272727272727275);
	      weights3d(229) = 0.1088239338943121;
	      points3d(241).Init(0.04999999999999999, 0.04999999999999999, 0.75);
	      weights3d(241) = -0.3914588212678253;
	      points3d(253).Init(0.04999999999999999, 0.04999999999999999, 0.65);
	      weights3d(253) = -0.3914588212678253;
	      points3d(265).Init(0.04999999999999999, 0.04999999999999999, 0.55);
	      weights3d(265) = -0.3914588212678253;
	      points3d(277).Init(0.15, 0.15, 0.65);
	      weights3d(277) = -0.3914588212678253;
	      points3d(289).Init(0.25, 0.25, 0.45);
	      weights3d(289) = -0.3914588212678253;
	      points3d(301).Init(0.35, 0.35, 0.25);
	      weights3d(301) = -0.3914588212678253;
	      points3d(313).Init(0.15, 0.15, 0.45);
	      weights3d(313) = -0.3914588212678253;
	      points3d(325).Init(0.25, 0.25, 0.35);
	      weights3d(325) = -0.3914588212678253;
	      points3d(337).Init(0.05555555555555552, 0.05555555555555552, 0.722222222222222);
	      weights3d(337) = 0.5552429621632808;
	      points3d(349).Init(0.05555555555555552, 0.05555555555555552, 0.6111111111111109);
	      weights3d(349) = 0.5552429621632808;
	      points3d(361).Init(0.05555555555555552, 0.05555555555555552, 0.5);
	      weights3d(361) = 0.5552429621632808;
	      points3d(373).Init(0.1666666666666665, 0.1666666666666665, 0.6111111111111109);
	      weights3d(373) = 0.5552429621632808;
	      points3d(385).Init(0.388888888888889, 0.388888888888889, 0.1666666666666665);
	      weights3d(385) = 0.5552429621632808;
	      points3d(397).Init(0.277777777777778, 0.277777777777778, 0.388888888888889);
	      weights3d(397) = 0.5552429621632808;
	      points3d(409).Init(0.1666666666666665, 0.1666666666666665, 0.388888888888889);
	      weights3d(409) = 0.5552429621632808;
	      points3d(421).Init(0.0625, 0.0625, 0.6875);
	      weights3d(421) = -0.3949061315567584;
	      points3d(433).Init(0.0625, 0.0625, 0.5625);
	      weights3d(433) = -0.3949061315567584;
	      points3d(445).Init(0.1875, 0.1875, 0.5625);
	      weights3d(445) = -0.3949061315567584;
	      points3d(457).Init(0.07142857142857151, 0.07142857142857151, 0.642857142857143);
	      weights3d(457) = 0.1483680859279701;
	      points3d(469).Init(0.07142857142857151, 0.07142857142857151, 0.5);
	      weights3d(469) = 0.1483680859279701;
	      points3d(481).Init(0.2142857142857145, 0.2142857142857145, 0.5);
	      weights3d(481) = 0.1483680859279701;
	      points3d(493).Init(0.357142857142857, 0.357142857142857, 0.2142857142857145);
	      weights3d(493) = 0.1483680859279701;
	      points3d(505).Init(0.08333333333333348, 0.08333333333333348, 0.5833333333333335);
	      weights3d(505) = -0.0285531476926435;
	      points3d(517).Init(0.25, 0.25, 0.4166666666666665);
	      weights3d(517) = -0.0285531476926435;
	      points3d(529).Init(0.09999999999999998, 0.09999999999999998, 0.5);
	      weights3d(529) = 0.002532258209857625;
	      points3d(541).Init(0.1363636363636365, 0.2272727272727275, 0.5909090909090911);
	      weights3d(541) = 0.1088239338943121;
	      points3d(565).Init(0.1363636363636365, 0.318181818181818, 0.5);
	      weights3d(565) = 0.1088239338943121;
	      points3d(589).Init(0.2272727272727275, 0.318181818181818, 0.409090909090909);
	      weights3d(589) = 0.1088239338943121;
	      points3d(613).Init(0.15, 0.25, 0.55);
	      weights3d(613) = -0.3914588212678253;
	      points3d(637).Init(0.15, 0.35, 0.45);
	      weights3d(637) = -0.3914588212678253;
	      points3d(661).Init(0.1666666666666665, 0.277777777777778, 0.5);
	      weights3d(661) = 0.5552429621632808;
	      points3d(685).Init(0.1875, 0.3125, 0.4375);
	      weights3d(685) = -0.3949061315567584;
	    } 
	    break; 
	  case 21 : 
	    { 
	      n1 = 1; n4 = 21; n6 = 8; n12 = 47; n24 = 12;
	      nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
	      points3d.Reallocate(nb_points); 
	      weights3d.Reallocate(nb_points); 
	      points3d(0).Init(0.25, 0.25, 0.25);
	      weights3d(0) = 0.8991968551558637;
	      points3d(1).Init(0.04166666666666652, 0.04166666666666652, 0.875);
	      weights3d(1) = 0.1482963604414264;
	      points3d(5).Init(0.2916666666666665, 0.2916666666666665, 0.1250000000000004);
	      weights3d(5) = 0.1482963604414264;
	      points3d(9).Init(0.2083333333333335, 0.2083333333333335, 0.375);
	      weights3d(9) = 0.1482963604414264;
	      points3d(13).Init(0.04545454545454553, 0.04545454545454553, 0.8636363636363635);
	      weights3d(13) = -0.572508521791816;
	      points3d(17).Init(0.318181818181818, 0.318181818181818, 0.04545454545454597);
	      weights3d(17) = -0.572508521791816;
	      points3d(21).Init(0.1363636363636365, 0.1363636363636365, 0.5909090909090911);
	      weights3d(21) = -0.572508521791816;
	      points3d(25).Init(0.2272727272727275, 0.2272727272727275, 0.318181818181818);
	      weights3d(25) = -0.572508521791816;
	      points3d(29).Init(0.04999999999999999, 0.04999999999999999, 0.85);
	      weights3d(29) = 0.8896791392450574;
	      points3d(33).Init(0.15, 0.15, 0.55);
	      weights3d(33) = 0.8896791392450574;
	      points3d(37).Init(0.05555555555555552, 0.05555555555555552, 0.8333333333333335);
	      weights3d(37) = -0.7138838084956467;
	      points3d(41).Init(0.277777777777778, 0.277777777777778, 0.166666666666666);
	      weights3d(41) = -0.7138838084956467;
	      points3d(45).Init(0.0625, 0.0625, 0.8125);
	      weights3d(45) = 0.3159249052454067;
	      points3d(49).Init(0.3125, 0.3125, 0.0625);
	      weights3d(49) = 0.3159249052454067;
	      points3d(53).Init(0.1875, 0.1875, 0.4375);
	      weights3d(53) = 0.3159249052454067;
	      points3d(57).Init(0.07142857142857151, 0.07142857142857151, 0.7857142857142855);
	      weights3d(57) = -0.0765266969523215;
	      points3d(61).Init(0.2142857142857145, 0.2142857142857145, 0.357142857142857);
	      weights3d(61) = -0.0765266969523215;
	      points3d(65).Init(0.08333333333333348, 0.08333333333333348, 0.75);
	      weights3d(65) = 0.009517715897547876;
	      points3d(69).Init(0.09999999999999998, 0.09999999999999998, 0.7);
	      weights3d(69) = -0.000531987018877625;
	      points3d(73).Init(0.3, 0.3, 0.1);
	      weights3d(73) = -0.000531987018877625;
	      points3d(77).Init(0.125, 0.125, 0.625);
	      weights3d(77) = 0.1483067872090884;
	      points3d(81).Init(0.1666666666666665, 0.1666666666666665, 0.5);
	      weights3d(81) = -0.7138838525829668;
	      points3d(85).Init(0.04166666666666652, 0.04166666666666652, 0.4583333333333335);
	      weights3d(85) = 0.1482963604414264;
	      points3d(91).Init(0.2083333333333335, 0.2083333333333335, 0.2916666666666665);
	      weights3d(91) = 0.1482963604414264;
	      points3d(97).Init(0.04999999999999999, 0.04999999999999999, 0.45);
	      weights3d(97) = 0.8896791392450574;
	      points3d(103).Init(0.15, 0.15, 0.35);
	      weights3d(103) = 0.8896791392450574;
	      points3d(109).Init(0.0625, 0.0625, 0.4375);
	      weights3d(109) = 0.3159249052454067;
	      points3d(115).Init(0.1875, 0.1875, 0.3125);
	      weights3d(115) = 0.3159249052454067;
	      points3d(121).Init(0.08333333333333348, 0.08333333333333348, 0.4166666666666665);
	      weights3d(121) = 0.009517715897547876;
	      points3d(127).Init(0.125, 0.125, 0.375);
	      weights3d(127) = 0.1483067872090884;
	      points3d(133).Init(0.04166666666666652, 0.04166666666666652, 0.7916666666666665);
	      weights3d(133) = 0.1482963604414264;
	      points3d(145).Init(0.04166666666666652, 0.04166666666666652, 0.7083333333333335);
	      weights3d(145) = 0.1482963604414264;
	      points3d(157).Init(0.04166666666666652, 0.04166666666666652, 0.625);
	      weights3d(157) = 0.1482963604414264;
	      points3d(169).Init(0.04166666666666652, 0.04166666666666652, 0.5416666666666665);
	      weights3d(169) = 0.1482963604414264;
	      points3d(181).Init(0.125, 0.125, 0.7083333333333335);
	      weights3d(181) = 0.1482963604414264;
	      points3d(193).Init(0.2083333333333335, 0.2083333333333335, 0.5416666666666665);
	      weights3d(193) = 0.1482963604414264;
	      points3d(205).Init(0.375, 0.375, 0.2083333333333335);
	      weights3d(205) = 0.1482963604414264;
	      points3d(217).Init(0.2916666666666665, 0.2916666666666665, 0.375);
	      weights3d(217) = 0.1482963604414264;
	      points3d(229).Init(0.125, 0.125, 0.5416666666666665);
	      weights3d(229) = 0.1482963604414264;
	      points3d(241).Init(0.125, 0.125, 0.4583333333333335);
	      weights3d(241) = 0.1482963604414264;
	      points3d(253).Init(0.2083333333333335, 0.2083333333333335, 0.4583333333333335);
	      weights3d(253) = 0.1482963604414264;
	      points3d(265).Init(0.04545454545454553, 0.04545454545454553, 0.7727272727272725);
	      weights3d(265) = -0.572508521791816;
	      points3d(277).Init(0.04545454545454553, 0.04545454545454553, 0.681818181818182);
	      weights3d(277) = -0.572508521791816;
	      points3d(289).Init(0.04545454545454553, 0.04545454545454553, 0.5909090909090911);
	      weights3d(289) = -0.572508521791816;
	      points3d(301).Init(0.04545454545454553, 0.04545454545454553, 0.5);
	      weights3d(301) = -0.572508521791816;
	      points3d(313).Init(0.1363636363636365, 0.1363636363636365, 0.681818181818182);
	      weights3d(313) = -0.572508521791816;
	      points3d(325).Init(0.409090909090909, 0.409090909090909, 0.1363636363636365);
	      weights3d(325) = -0.572508521791816;
	      points3d(337).Init(0.2272727272727275, 0.2272727272727275, 0.5);
	      weights3d(337) = -0.572508521791816;
	      points3d(349).Init(0.1363636363636365, 0.1363636363636365, 0.5);
	      weights3d(349) = -0.572508521791816;
	      points3d(361).Init(0.1363636363636365, 0.1363636363636365, 0.409090909090909);
	      weights3d(361) = -0.572508521791816;
	      points3d(373).Init(0.2272727272727275, 0.2272727272727275, 0.409090909090909);
	      weights3d(373) = -0.572508521791816;
	      points3d(385).Init(0.318181818181818, 0.318181818181818, 0.2272727272727275);
	      weights3d(385) = -0.572508521791816;
	      points3d(397).Init(0.04999999999999999, 0.04999999999999999, 0.75);
	      weights3d(397) = 0.8896791392450574;
	      points3d(409).Init(0.04999999999999999, 0.04999999999999999, 0.65);
	      weights3d(409) = 0.8896791392450574;
	      points3d(421).Init(0.04999999999999999, 0.04999999999999999, 0.55);
	      weights3d(421) = 0.8896791392450574;
	      points3d(433).Init(0.15, 0.15, 0.65);
	      weights3d(433) = 0.8896791392450574;
	      points3d(445).Init(0.25, 0.25, 0.45);
	      weights3d(445) = 0.8896791392450574;
	      points3d(457).Init(0.35, 0.35, 0.25);
	      weights3d(457) = 0.8896791392450574;
	      points3d(469).Init(0.15, 0.15, 0.45);
	      weights3d(469) = 0.8896791392450574;
	      points3d(481).Init(0.25, 0.25, 0.35);
	      weights3d(481) = 0.8896791392450574;
	      points3d(493).Init(0.05555555555555552, 0.05555555555555552, 0.722222222222222);
	      weights3d(493) = -0.7138838084956467;
	      points3d(505).Init(0.05555555555555552, 0.05555555555555552, 0.6111111111111109);
	      weights3d(505) = -0.7138838084956467;
	      points3d(517).Init(0.05555555555555552, 0.05555555555555552, 0.5);
	      weights3d(517) = -0.7138838084956467;
	      points3d(529).Init(0.1666666666666665, 0.1666666666666665, 0.6111111111111109);
	      weights3d(529) = -0.7138838084956467;
	      points3d(541).Init(0.388888888888889, 0.388888888888889, 0.1666666666666665);
	      weights3d(541) = -0.7138838084956467;
	      points3d(553).Init(0.277777777777778, 0.277777777777778, 0.388888888888889);
	      weights3d(553) = -0.7138838084956467;
	      points3d(565).Init(0.1666666666666665, 0.1666666666666665, 0.388888888888889);
	      weights3d(565) = -0.7138838084956467;
	      points3d(577).Init(0.0625, 0.0625, 0.6875);
	      weights3d(577) = 0.3159249052454067;
	      points3d(589).Init(0.0625, 0.0625, 0.5625);
	      weights3d(589) = 0.3159249052454067;
	      points3d(601).Init(0.1875, 0.1875, 0.5625);
	      weights3d(601) = 0.3159249052454067;
	      points3d(613).Init(0.07142857142857151, 0.07142857142857151, 0.642857142857143);
	      weights3d(613) = -0.0765266969523215;
	      points3d(625).Init(0.07142857142857151, 0.07142857142857151, 0.5);
	      weights3d(625) = -0.0765266969523215;
	      points3d(637).Init(0.2142857142857145, 0.2142857142857145, 0.5);
	      weights3d(637) = -0.0765266969523215;
	      points3d(649).Init(0.357142857142857, 0.357142857142857, 0.2142857142857145);
	      weights3d(649) = -0.0765266969523215;
	      points3d(661).Init(0.08333333333333348, 0.08333333333333348, 0.5833333333333335);
	      weights3d(661) = 0.009517715897547876;
	      points3d(673).Init(0.25, 0.25, 0.4166666666666665);
	      weights3d(673) = 0.009517715897547876;
	      points3d(685).Init(0.09999999999999998, 0.09999999999999998, 0.5);
	      weights3d(685) = -0.000531987018877625;
	      points3d(697).Init(0.125, 0.2083333333333335, 0.625);
	      weights3d(697) = 0.1482963604414264;
	      points3d(721).Init(0.125, 0.2916666666666665, 0.5416666666666665);
	      weights3d(721) = 0.1482963604414264;
	      points3d(745).Init(0.125, 0.375, 0.4583333333333335);
	      weights3d(745) = 0.1482963604414264;
	      points3d(769).Init(0.2083333333333335, 0.2916666666666665, 0.4583333333333335);
	      weights3d(769) = 0.1482963604414264;
	      points3d(793).Init(0.2083333333333335, 0.2916666666666665, 0.375);
	      weights3d(793) = 0.1482963604414264;
	      points3d(817).Init(0.1363636363636365, 0.2272727272727275, 0.5909090909090911);
	      weights3d(817) = -0.572508521791816;
	      points3d(841).Init(0.1363636363636365, 0.318181818181818, 0.5);
	      weights3d(841) = -0.572508521791816;
	      points3d(865).Init(0.2272727272727275, 0.318181818181818, 0.409090909090909);
	      weights3d(865) = -0.572508521791816;
	      points3d(889).Init(0.15, 0.25, 0.55);
	      weights3d(889) = 0.8896791392450574;
	      points3d(913).Init(0.15, 0.35, 0.45);
	      weights3d(913) = 0.8896791392450574;
	      points3d(937).Init(0.1666666666666665, 0.277777777777778, 0.5);
	      weights3d(937) = -0.7138838084956467;
	      points3d(961).Init(0.1875, 0.3125, 0.4375);
	      weights3d(961) = 0.3159249052454067;
	    } 
	    break; 
	  default :
	    formula_not_available = true;
	  }
      }
    else if (type_quad == QUADRATURE_MASS_LUMPED)
      {
        switch (p)
          {
          case 0 :
            n1 = 0; n4 = 1; n6 = 0; n12 = 0; n24 = 0;
            nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
            points3d.Reallocate(nb_points); 
            weights3d.Reallocate(nb_points); 
            points3d(0).Init(0.0, 0.0, 0.0);
            weights3d(0) = 1.0/24;
            break;
          default :
            {
              cout << "Order not implemented " << endl;
              abort();
            }
          }
      }
    else if (type_quad == QUADRATURE_QUASI_LUMPED)
      {
        Real_wp one(1);
        switch (p)
          {
          case 0 :
            n1 = 0; n4 = 1; n6 = 0; n12 = 0; n24 = 0;
            nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
            points3d.Reallocate(nb_points); 
            weights3d.Reallocate(nb_points); 
            points3d(0).Init(0.0, 0.0, 0.0);
            weights3d(0) = one/24;
            break;
          case 2 :
            n1 = 1; n4 = 1; n6 = 1; n12 = 0; n24 = 0;
            nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
            points3d.Reallocate(nb_points); 
            weights3d.Reallocate(nb_points); 
            points3d(0).Init(0.25, 0.25, 0.25);
            weights3d(0) = 8.0/9;
            points3d(1).Init(0.0, 0.0, 0.0);
            weights3d(1) = 4.0/90;
            points3d(5).Init(0.5, 0.0, 0.0);
            weights3d(5) = 4.0/90;
            Mlt(one/8, weights3d);
            break;
          case 4 :
            n1 = 0; n4 = 3; n6 = 0; n12 = 1; n24 = 0;
            nb_points = n1 + 4*n4 + 6*n6 + 12*n12 + 24*n24;
            points3d.Reallocate(nb_points); 
            weights3d.Reallocate(nb_points); 
            points3d(0).Init(0.0, 0.0, 0.0);
            weights3d(0) = 2.104389501311755e-03;            
            points3d(4).Init(1.0/3, 1.0/3, 0.0);
            weights3d(4) = 7.266253849264000e-2;
            points3d(8).Init(0.155051025721693, 0.155051025721693, 5.348469228349211e-01);
            weights3d(8) = 1.990425958155520e-01;
            points3d(12).Init(0.000000000000000, 0.276393202250021, 0.723606797749979);
            weights3d(12) = 1.984126984127200e-02;
            Mlt(one/8, weights3d);
            break;
          default :
            {
              cout << "Order not implemented " << endl;
              abort();
            }
          }
      }
    else
      formula_not_available = true;
    
    
    if (formula_not_available)
      {
        int r = p/2;
        GetTensorizedQuadratureRule(r, points3d, weights3d, type_quad);
      }
    else
      {
        int nb_couples = n1 + n4 + n6 + n12 + n24;
        IVect NbPoints_couple(nb_couples);
        int ind = 0;
        for (int i = 0; i < n1; i++)
          NbPoints_couple(ind++) = 1;
        
        for (int i = 0; i < n4; i++)
          NbPoints_couple(ind++) = 4;
        
        for (int i = 0; i < n6; i++)
          NbPoints_couple(ind++) = 6;
        
        for (int i = 0; i < n12; i++)
          NbPoints_couple(ind++) = 12;
        
        for (int i = 0; i < n24; i++)
          NbPoints_couple(ind++) = 24;
        
        CompleteTetrahedronPointsWithSymmetry(points3d, weights3d, nb_couples, NbPoints_couple);
      }
  }
  
  
  void TetrahedronQuadrature
  ::GetTensorizedQuadratureRule(int r, VectR3& points3d, VectReal_wp& weights3d, int type_quad)
  {
    // quadrature points based on Gauss-Jacobi rules (with tensorization)
    VectReal_wp PointsX, PointsY, PointsZ, WeightsX, WeightsY, WeightsZ;
    ComputeGaussJacobi(PointsZ, WeightsZ, r, Real_wp(2), Real_wp(0));
    ComputeGaussJacobi(PointsY, WeightsY, r, Real_wp(1), Real_wp(0));
    ComputeGaussLegendre(PointsX, WeightsX, r);
    
    points3d.Reallocate((r+1)*(r+1)*(r+1));
    weights3d.Reallocate((r+1)*(r+1)*(r+1));
    int node = 0;
    for (int i = 0; i <= r; i++)
      for (int j = 0; j <= r; j++)
	for (int k = 0; k <= r; k++)
	  {
	    Real_wp x = PointsX(i), y = PointsY(j), z = PointsZ(k);
	    points3d(node).Init((1.0-z)*(1.0-y)*x, (1.0-z)*y, z);
	    weights3d(node) = WeightsX(i)*WeightsY(j)*WeightsZ(k);
            node++;
	  }
  }
  
  
  //! quadrature points are generated from original points by symmetry considerations
  void TetrahedronQuadrature::
  CompleteTetrahedronPointsWithSymmetry(VectR3& points3d, VectReal_wp& weights3d,
                                        int nb_couples, const IVect& NbPoints_couple)
  {
    int ind = 0;
    // now we generate the other points by symmetry
    for (int i = 0; i < nb_couples; i++)
      {
	int np = NbPoints_couple(i);
	Real_wp x = points3d(ind)(0);
	Real_wp y = points3d(ind)(1);
	Real_wp z = points3d(ind)(2);
	Real_wp t = 1.0-x-y-z;
	
	if (np==4)
	  {
            if (abs(x-y) > R3::threshold)
              x = z;
            
            z = 1.0 - 3.0*x;
	    // ind is l1 l1 l2 l1
            points3d(ind).Init(x, x, z);
	    points3d(ind+1).Init(z, x, x);
	    points3d(ind+2).Init(x, z, x);
	    points3d(ind+3).Init(x, x, x);
	  }
	else if (np==6)
	  {
            z = 0.5 - x;
	    points3d(ind).Init(x, x, z);
	    points3d(ind+1).Init(z, x, x);
	    points3d(ind+2).Init(z, z, x);
	    points3d(ind+3).Init(z, x, z);
	    points3d(ind+4).Init(x, z, x);
	    points3d(ind+5).Init(x, z, z);
	  }
	else if (np==12)
	  {
            Real_wp tmp;
            if ((abs(y-z) <= R3::threshold)||(abs(y-t) <= R3::threshold))
              {
                tmp = x;
                x = y;
                y = tmp;
              }

            if (abs(z-t) <= R3::threshold)
              {
                tmp = x;
                x = z;
                z = tmp;
              }
              
              
	    points3d(ind).Init(x, x, z);
	    points3d(ind+1).Init(x, x, t);
	    points3d(ind+2).Init(x, z, x);
	    points3d(ind+3).Init(x, t, x);
	    points3d(ind+4).Init(z, x, x);
	    points3d(ind+5).Init(t, x, x);
	    points3d(ind+6).Init(x, t, z);
	    points3d(ind+7).Init(x, z, t);
	    points3d(ind+8).Init(t, x, z);
	    points3d(ind+9).Init(z, x, t);
	    points3d(ind+10).Init(t, z, x);
	    points3d(ind+11).Init(z, t, x);
	  }
	else if (np==24)
	  {
	    points3d(ind).Init(x, y, z);
	    points3d(ind+1).Init(x, y, t);
	    points3d(ind+2).Init(x, z, y);
	    points3d(ind+3).Init(x, z, t);
	    points3d(ind+4).Init(x, t, y);
	    points3d(ind+5).Init(x, t, z);
	    points3d(ind+6).Init(y, x, z);
	    points3d(ind+7).Init(y, x, t);
	    points3d(ind+8).Init(y, z, x);
	    points3d(ind+9).Init(y, z, t);
	    points3d(ind+10).Init(y, t, x); 
	    points3d(ind+11).Init(y, t, z);
	    points3d(ind+12).Init(z, x, y);
	    points3d(ind+13).Init(z, x, t);
	    points3d(ind+14).Init(z, y, x);
	    points3d(ind+15).Init(z, y, t);
	    points3d(ind+16).Init(z, t, x);
	    points3d(ind+17).Init(z, t, y);
	    points3d(ind+18).Init(t, x, y);
	    points3d(ind+19).Init(t, x, z);
	    points3d(ind+20).Init(t, y, x); 
	    points3d(ind+21).Init(t, y, z);
	    points3d(ind+22).Init(t, z, x);
	    points3d(ind+23).Init(t, z, y);
	  }
        
	for (int j = 1; j < np; j++)
	  weights3d(ind+j) = weights3d(ind);
	  
	ind += np;
      }
   
  }
    
}

#define MONTJOIE_FILE_TETRAHEDRON_QUADRATURE_CXX
#endif
